import numpy as np
from IPython.display import Math
try:
    from collections.abc import MutableMapping
except:
    from collections import MutableMapping
from sympy import symbols, S, binomial, summation, sqrt, cos, sin, Function,atan2,expand_trig,diff,Matrix, series
from .hamiltonian import Hamiltonian,PhaseSpaceState
from .miscellaneous import poisson_bracket
from .disturbing_function import  _p1_p2_from_k_nu, evaluate_df_coefficient_delta_expansion, get_df_term_latex
from .disturbing_function import list_resonance_terms, list_secular_terms, _nucombos, _lcombos
from .disturbing_function import k_nu_depend_on_eccentricities, k_nu_depend_on_inclinations
from .disturbing_function import df_coefficient_C,get_df_coefficient_symbol,evaluate_df_coefficient_delta_expansion
from .nbody_simulation_utilities import reb_add_poincare_particle, reb_calculate_orbits
from itertools import combinations
import rebound
import warnings
from IPython.display import Math
def get_re_im_components(x,y,k):
    """
    Get the real and imaginary components of
        (x + sgn(k) * i y)^|k|
    """
    if k==0:
        return 1,0
    absk = abs(k)
    sgnk = np.sign(k)
    re,im=0,0
    for l in range(0,absk+1):
        b = binomial(absk,l)
        if l%2==0:
            re += b * (sgnk * y)**l * x**(absk-l) * (-1)**(l//2)
        else:
            im += b * (sgnk * y)**l * x**(absk-l) * (-1)**((l-1)//2)
    return re,im

def num_passed(iterable): # returns num of entries in iterable that are not None
    ctr = 0
    for i in iterable:
        if i is not None:
            ctr += 1
    return ctr

def _get_a0_symbol(i):
    return symbols(r"a_{{{0}\,0}}".format(i),positive=True)

def _get_Lambda0_symbol(i):
    return symbols(r"\Lambda_{{{0}\,0}}".format(i),positive=True)

class PoincareParticle(object):
    """
    A class representing an individual member (star, planet, or test particle) of a planetary system.
    The appropriate value for mu and M depends on the adopted coordinate system and Kepler splitting
    (see e.g., Hernandez and Dehnen 2017 for a review and comparison). celmech supports canonical
    heliocentric coordinates (default) and democratic heliocentric coordinates.

    Parameters 
    ----------
    coordinates: str
      Specifices the canonical coordinate system. This determines the appropriate definitions of mu and M. Options:
      'canonical heliocentric' (default): canonical heliocentric coordinates in the COM frame e.g. Laskar & Robutel 1995
      'democratic heliocentric': e.g. Duncan et al. 1998
    G : float
      Gravitational constant (Default: 0)
    m : float
      Physical mass of particle.
    Mstar : float
      Physical mass of central body.
    mu : float
      'Canonical' mass of body. mu=reduced mass for canonical heliocentric coordinates (default)
      mu=m for democratic heliocentric coordinates.
    M : float
      'Canonical' central mass. M=Mstar+m for canonical heliocentric coordinates (default)
      M=Mstar for democratic heliocentric coordinates.
    Lambda, sLambda, a: float
      These variables specify the semimajor axis of the orbit. 
      Can pass any of the three, but at least one must be specified.
    l : float
      Mean longitude of the orbit. If not passed, defaults to 0.
    Gamma, sGamma, e: float 
      These variables specify the orbital eccentricity. Any one can be passed. If none passed, defaults to 0
    gamma, pomega: float
      These variables specify the pericenter orientation. Any one can be passed. If none passed, defaults to 0
    Q, sQ, inc: float 
      These variables specify the orbital inclination. Any one can be passed. If none passed, defaults to 0
    q, Omega: float
      These variables specify the node longitude. Any one can be passed. If none passed, defaults to 0
    """
    def __init__(self, coordinates='canonical heliocentric', G=1., m=None, Mstar=None, mu=None, M=None, sLambda=None, l=None, sGamma=None, gamma=None, sQ=None, q=None, Lambda=None, Gamma=None, Q=None, a=None, e=None, inc=None, pomega=None, Omega=None, skappa=None, seta=None, ssigma=None, srho=None, kappa=None, eta=None, sigma=None, rho=None):
        """
        We store Cartesian components of specific actions to support test particles
        """
        self.G = G  
        self.coordinates = coordinates

        massError = False
        if m is not None and Mstar is not None: # passed both physical masses
            if mu is not None or M is not None: # also passed one of others
                massError = True
            else: # calculate from physical masses
                if coordinates == 'democratic heliocentric':
                    self.mu = m
                    self.M = Mstar
                elif coordinates == 'canonical heliocentric':
                    self.mu = m*Mstar/(Mstar + m)
                    self.M = Mstar+m
                else:
                    raise AttributeError("coordinates must either be 'canonical heliocentric' (default) or 'democratic heiocentric")
        else: # didn't pass both physical masses
            if m is not None or Mstar is not None: # passed only one physical mass
                massError = True
            elif mu is None or M is None: # didn't pass both can. masses
                massError = True
            else:
                self.mu = mu
                self.M = M

        if massError == True:   
            raise AttributeError("Have to either pass physical masses (m, Mstar) or 'canonical masses' (mu and M). Can't mix or pass both.")

        # if the variables we actually store are passed then we save them and are done
        num = num_passed([sLambda, l, skappa, seta, ssigma, srho])
        if num == 6:
            self.sLambda = sLambda
            self.l = l
            self.skappa = skappa
            self.seta = seta
            self.ssigma = ssigma
            self.srho = srho
            return
        
        # if passed our massive variables, just normalize by mu
        num = num_passed([Lambda, l, kappa, eta, sigma, rho])
        if num == 6:
            self.sLambda = Lambda/self.mu
            self.l = l
            self.skappa = kappa/np.sqrt(self.mu) # these are all sqrt(Gamma) variables so norm by sqrt(mass)
            self.seta = eta/np.sqrt(self.mu)
            self.ssigma = sigma/np.sqrt(self.mu)
            self.srho = rho/np.sqrt(self.mu)
            return

        # need to calculate the variables we store
        num = num_passed([sLambda, Lambda, a])
        if num == 0:
            raise AttributeError("Must pass exactly 1 of Lambda, sLambda (specific Lambda, i.e., per unit mass), and a.")
        elif num > 1:
            raise AttributeError("Can't pass more than 1 of Lambda, sLambda (specific Gamma, i.e. per unit mass), and a (semimajor axis)")
        num = num_passed([sGamma, Gamma, e])
        if num == 0:
            sGamma = 0 # default
        elif num > 1:
            raise AttributeError("Can't pass more than 1 of Gamma, sGamma (specific Gamma, i.e. per unit mass), and e (eccentricity)")
        num = num_passed([sQ, Q, inc])
        if num == 0:
            sQ = 0 # default
        elif num > 1:
            raise AttributeError("Can't pass more than 1 of Q, sQ (specific Q, i.e. per unit mass), and inc (inclination)")
        num = num_passed([gamma, pomega])
        if num == 0:
            gamma = 0 # default
        elif num > 1:
            raise AttributeError("Can't pass more than 1 of gamma (-pomega) and pomega (longitude of pericenter)")
        num = num_passed([q, Omega])
        if num == 0:
            q = 0 # default
        elif num > 1:
            raise AttributeError("Can't pass more than 1 of q (-Omega) and Omega (longitude of ascending node)")
        num = num_passed([l])
        if num == 0:
            l = 0 # default
       
        self.l = l

        if pomega is not None:
            gamma = -pomega
        if Omega is not None:
            q = -Omega

        if sLambda is not None:
            self.sLambda = sLambda
        elif Lambda is not None:
            try:
                self.sLambda = Lambda/self.mu
            except:
                raise AttributeError("Need to pass specific actions (sLambda, sGamma, and sQ) or a, e, and inc for test particles")
        elif a is not None:
            self.sLambda = np.sqrt(self.G*self.M*a)

        if Gamma is not None:
            try:
                sGamma = Gamma/self.mu
            except:
                raise AttributeError("Need to pass specific actions (sLambda, sGamma, and sQ) or a, e, and inc for test particles")
        elif e is not None:
            sGamma = self.sLambda*(1.-np.sqrt(1.-e**2))

        if Q is not None:
            try:
                sQ = Q/self.mu
            except:
                raise AttributeError("Need to pass specific actions (sLambda, sGamma, and sQ) or a, e, and inc for test particles")
        elif inc is not None:
            sQ = (self.sLambda - sGamma) * (1 - np.cos(inc))

        self.skappa = np.sqrt(2.*sGamma)*np.cos(gamma) # X per unit sqrt(mass)
        self.seta = np.sqrt(2.*sGamma)*np.sin(gamma)

        self.ssigma = np.sqrt(2.*sQ)*np.cos(q) # Xinc per unit sqrt(mass)
        self.srho = np.sqrt(2.*sQ)*np.sin(q)

    def _mu_M_to_m_Mstar(self, mu, M):
        """
        Takes reduced mass mu = mMstar/(Mstar+m) and M=Mstar+m
        and returns m and Mstar 
        """
        d = np.sqrt(M**2 - 4*mu*M) # m_0 - m_i
        Mstar = (M+d)/2
        m = mu*M/Mstar
        return m, Mstar
   
    @property
    def m(self):
        if self.coordinates == "democratic heliocentric":
            return self.mu
        elif self.coordinates == "canonical heliocentric":
            m, Mstar = self._mu_M_to_m_Mstar(self.mu, self.M)
            return m
    @m.setter
    def m(self, value):
        if self.coordinates == "democratic heliocentric":
            self.mu = value
        elif self.coordinates == "canonical heliocentric":
            raise AttributeError("Can't change physical masses after initialization with canonical heliocentric coordinates.")
    @property
    def Mstar(self):
        if self.coordinates == "democratic heliocentric":
            return self.M
        elif self.coordinates == "canonical heliocentric":
            m, Mstar = self._mu_M_to_m_Mstar(self.mu, self.M)
            return Mstar
    @Mstar.setter
    def Mstar(self, value):
        if self.coordinates == "democratic heliocentric":
            self.M = value
        elif self.coordinates == "canonical heliocentric":
            raise AttributeError("Can't change physical masses after initialization with canonical heliocentric coordinates.")
    @property
    def x(self):
        return (self.kappa - 1j * self.eta) / np.sqrt(2)
    @property
    def X(self):
        return self.x * np.sqrt(2 / self.Lambda)
    @property
    def y(self):
        return (self.sigma - 1j * self.rho) / np.sqrt(2)
    @property
    def Y(self):
        return self.y * np.sqrt(0.5 / self.Lambda)

    @property
    def xbar(self):
        return np.conj(self.x)
    @property
    def Xbar(self):
        return np.conj(self.X)
    @property
    def ybar(self):
        return np.conj(self.y)
    @property
    def Ybar(self):
        return np.conj(self.Y)

    @property
    def kappa(self):
        return np.sqrt(self.mu)*self.skappa
    @kappa.setter
    def kappa(self, value):
        self.skappa = value/np.sqrt(self.mu)
    @property
    def eta(self):
        return np.sqrt(self.mu)*self.seta
    @eta.setter
    def eta(self, value):
        self.seta = value/np.sqrt(self.mu)

    @property
    def sigma(self):
        return np.sqrt(self.mu)*self.ssigma
    @sigma.setter
    def sigma(self, value):
        self.ssigma = value/np.sqrt(self.mu)

    @property
    def rho(self):
        return np.sqrt(self.mu)*self.srho
    @rho.setter
    def rho(self, value):
        self.srho = value/np.sqrt(self.mu)

    @property
    def Lambda(self):
        return self.mu*self.sLambda
    @Lambda.setter
    def Lambda(self, value):
        self.sLambda = value/self.mu

    @property
    def Gamma(self):
        return self.mu*(self.skappa**2+self.seta**2)/2.
    @Gamma.setter
    def Gamma(self, value):
        self.sGamma = value/self.mu

    @property
    def Q(self):
        return self.mu*(self.ssigma**2+self.srho**2)/2.
    @Q.setter
    def Q(self, value):
        self.sQ = value/self.mu

    @property
    def sGamma(self):
        return (self.skappa**2+self.seta**2)/2.
    @property
    def gamma(self):
        return np.arctan2(self.seta, self.skappa)

    @property
    def sQ(self):
        return (self.ssigma**2+self.srho**2)/2.
    @property
    def q(self):
        return np.arctan2(self.srho,self.ssigma)

    @property
    def a(self):
        return self.sLambda**2/self.G/self.M
    @property
    def n(self):
        return self.G**2*self.M**2/self.sLambda**3
    @property
    def P(self):
        return 2 * np.pi / self.n
    @property
    def e(self):
        GbyL = self.sGamma/self.sLambda
        if 1-(1.-GbyL)*(1.-GbyL) < 0:
            raise AttributeError("sGamma:{0}, sLambda:{1}, GbyL:{2}, val:{3}".format(self.sGamma, self.sLambda, GbyL, 1-(1.-GbyL)*(1.-GbyL)))
        return np.sqrt(1 - (1-GbyL)*(1-GbyL))
    @property
    def inc(self):
        QbyLminusG = self.sQ / (self.sLambda - self.sGamma)
        cosi = 1 - QbyLminusG
        if np.abs(cosi) > 1:
            raise AttributeError("sGamma:{0}, sLambda:{1}, sQ:{2}, cosi:{3}".format(self.sGamma, self.sLambda, self.sQ,cosi))
        return np.arccos(cosi)
    @property
    def pomega(self):
        return -self.gamma
    @property
    def Omega(self):
        return -self.q

    def __repr__(self):
        """ 
        Returns a string with the state of the particle.
        """ 
        return '<{0}.{1} object, mu={2} M={3} sLambda={4} l={5} skappa={6} seta={7} srho={8} ssigma={9}>'.format(self.__module__, type(self).__name__, self.mu, self.M, self.sLambda, self.l, self.skappa, self.seta, self.srho, self.ssigma)

class PoincareParticles(MutableMapping):
    """
    """
    def __init__(self, poincare):
        self.poincare = poincare 

    def __getitem__(self, i):
        # go from int key and generate a PoincareParticle
        # need G and masses
        if i == 0:
            return PoincareParticle(G=np.nan, m=np.nan, Mstar=np.nan, l=np.nan, eta=np.nan, rho=np.nan, Lambda=np.nan, kappa=np.nan, sigma=np.nan)    #raise AttributeError("No Poincare elements for the central star")
        p = self.poincare
        if isinstance(i, slice):
            return [self[i] for i in range(*i.indices(p.N))]

        if i < 0: # accept negative indices
            i += p.N
        if i < 0 or i >= p.N:
            raise AttributeError("Index {0} used to access particles out of range.".format(i))
   
        if p.masses[i] == 0: 
            raise AttributeError("Current implementation of Poincare does not work with test particles")

        val = p.values
        j = i-1 # index starting at 0 instead of 1
        l = val[j * 3]
        eta = val[j * 3 + 1]
        rho = val[j * 3 + 2]
        Lambda = val[p.N_dof + j * 3]
        kappa = val[p.N_dof + j * 3 + 1]
        sigma = val[p.N_dof + j * 3 + 2]

        return PoincareParticle(coordinates=p.coordinates, G=p.G, m=p.masses[i], Mstar=p.masses[0], l=l, eta=eta, rho=rho, Lambda=Lambda, kappa=kappa, sigma=sigma)

    def __setitem__(self, key, value):
        # we could allow user to set only the stored variables sLambda, l, skappa etc.
        raise AttributeError("Can't set Poincare particle attributes")

    def __delitem__(self, key):
        raise AttributeError("deleting variables not implemented.")

    def __iter__(self):
        for p in self[:self.poincare.N]:
            yield p

    def __len__(self):
        return self.poincare.N

# Poincare is a phasespacestate with G, masses and coordinates
class Poincare(PhaseSpaceState):
    """
    A class representing a collection of Poincare particles constituting a planetary system.
    """
    def __init__(self, G, poincareparticles, coordinates="canonical heliocentric",t=0):
        # additional variables that need storing in addition to phasespacestate variables
        self.G = G
        self.masses = [poincareparticles[0].Mstar] + [p.m for p in poincareparticles]
        self.coordinates = coordinates
        initial_p_values = []
        initial_q_values = []
        qvars = []
        pvars = []

        for i,particle in enumerate(poincareparticles):
            q = list(symbols("lambda{0}, eta{0}, rho{0}".format(i+1)))
            p = list(symbols("Lambda{0}, kappa{0}, sigma{0}".format(i+1)))
            pvals = [particle.Lambda,particle.kappa,particle.sigma]
            qvals = [particle.l,particle.eta,particle.rho]
            initial_p_values += pvals
            initial_q_values += qvals
            qvars += q
            pvars += p
        qpvars = qvars + pvars
        initial_values = initial_q_values + initial_p_values
        super(Poincare,self).__init__(qpvars,initial_values,t=t) 
    
    @property
    def N(self):
        """
        Return total number of bodies, including central star (analogous to REBOUND sim.N)
        """
        return int(self.N_dof/3) + 1 # +1 for star

    @property
    def particles(self):
        particles = PoincareParticles(self)
        return particles

    @classmethod
    def from_Simulation(cls, sim, coordinates="canonical heliocentric"):
        """ 
        Convert REBOUND Simulation to Poincare object, using specified canonical coordinates.
        Assumes the dominant mass is sim.particles[0].

        Parameters 
        ----------
        
        sim : rebound.Simulation
          Simulation to convert.
        coordinates: str
          Specifices the canonical coordinate system. This determines the appropriate definitions of mu and M. Options:
          'canonical heliocentric' (default): canonical heliocentric coordinates in the COM frame e.g. Laskar & Robutel 1995
          'democratic heliocentric': e.g. Duncan et al. 1998
    
        Returns
        -------

        Poincare object
        """ 
        sim = sim.copy()
        # Move to COM frame so P0 = 0 in canonical heliocentric coordinates
        sim.move_to_com()
        particles = []
        #pvars = Poincare(G=sim.G, coordinates=coordinates,t=sim.t)
        o = reb_calculate_orbits(sim, coordinates=coordinates)
        for i in range(1,sim.N_real):
            orb = o[i-1]
            if orb.a <= 0. or orb.e >= 1.:
                raise AttributeError("Celmech error: Poincare.from_Simulation only support elliptical orbits. Particle {0}'s (heliocentric) a={1}, e={2}".format(i, orb.a, orb.e))
            # always pass physical masses, PoincareParticle will calculate appropriate canonical mass based on coord
            particle = PoincareParticle(m=sim.particles[i].m, Mstar=sim.particles[0].m, a=orb.a, l=orb.l, e=orb.e, pomega=orb.pomega, inc=orb.inc, Omega=orb.Omega,coordinates=coordinates,G=sim.G)
            particles.append(particle)
        return cls(G=sim.G,poincareparticles=particles, coordinates=coordinates,t=sim.t)

    def to_Simulation(self):
        """ 
        Convert Poincare object to a REBOUND simulation in COM frame.

        Returns
        -------
        sim : rebound.Simulation
        """ 

        sim = rebound.Simulation()
        sim.G = self.G
        sim.t = self.t
        ps = self.particles
        Mstar = ps[1].Mstar # use first Poincare particle to extract Mstar
        sim.add(m=Mstar)
        for i in range(1, self.N):
            reb_add_poincare_particle(ps[i], sim)
        return sim
    
    def copy(self):
        return Poincare(G=self.G, coordinates=self.coordinates, poincareparticles=self.particles[1:self.N],t=self.t)

# If we wanted Poincare.from_Hamiltonian, would need PoincareHamiltonian to hold both m_0 (for star) and coordinates
class PoincareHamiltonian(Hamiltonian):
    """
    A class representing the Hamiltonian governing the dynamical evolution of a system of particles,
    stored as a :class:`celmech.poincare.Poincare` instance.

    Attributes
    ----------
    H : sympy expression
        Symbolic expression for the Hamiltonian.
    NH : sympy expression
        Symbolic expression for the Hamiltonian with 
        numerical values of parameters substituted
        where applicable.
    N : int
        Number of particles
    particles : list
        List of :class:`celmech.poincare.PoincareParticle`s 
        making up the system.
    state : :class:`celmech.poincare.Poincare`
        A set of Poincare variables to which 
        transformations are applied.
    """
    def __init__(self, pvars):
        H_params = {symbols('G'):pvars.G}
        ps = pvars.particles
        H = S(0) 
        self.Lambda0s = [None] + [_get_Lambda0_symbol(i) for i in range(1,pvars.N)]
        for i in range(1, pvars.N):
            H_params[symbols("mu{0}".format(i))] = ps[i].mu
            H_params[symbols("m{0}".format(i))] = ps[i].m
            H_params[symbols("M{0}".format(i))] = ps[i].M
            H_params[self.Lambda0s[i]] = ps[i].Lambda
            H_params[_get_a0_symbol(i)] = ps[i].a
            for j in range(i+1,pvars.N):
                alpha_sym = symbols(r"\alpha_{{{0}\,{1}}}".format(i,j))
                alpha_val = ps[i].a/ps[j].a
                H_params[alpha_sym] = alpha_val

            H = self.add_Hkep_term(H, i)
        self.resonance_indices = []
        super(PoincareHamiltonian, self).__init__(H, H_params, pvars) 
    
    @property
    def particles(self):
        return self.state.particles

    @property
    def N(self):
        return len(self.particles)

    @property 
    def t(self):
        return self.state.t

    @property
    def df_raw_latex(self):
        """
        Get the raw latex string (r"...") for the terms currently included in the disturbing function.
        This property is likely not what you're looking for.
        Most users will either want df to pretty print df in a jupyter notebook,
        or df_latex for the actual latex (with single slashes) to copy-paste into a paper.
        """
        d = r""
        for i1, i2, (kvec, nuvec, lvec) in self.resonance_indices:
            d += get_df_term_latex(*kvec,*nuvec,*lvec,i1,i2)
        return d
    
    @property
    def df(self):
        """
        Pretty print the terms currently included in the disturbing function in a jupyter notebook.
        """
        display(Math(self.df_raw_latex))
    
    @property
    def df_latex(self):
        """
        Print the latex for the terms currently included in the disturbing function
        (e.g., to copy-paste into a paper)
        """
        print(self.df_raw_latex)
    
    def add_Hkep_term(self, H, index):
        """
        Add the Keplerian component of the Hamiltonian for planet ''.
        """
        G, M, mu, Lambda = symbols('G, M{0}, mu{0}, Lambda{0}'.format(index))
        #m, M, mu, Lambda, lam, Gamma, gamma = self._get_symbols(index)
        H +=  -G**2*M**2*mu**3 / (2 * Lambda**2)
        return H

    def add_cosine_term(self,k_vec,max_order=None,l_max=0,nu_vecs=None,l_vecs=None,indexIn=1,indexOut=2,eccentricities=True,inclinations=True):
        """
        Add a specific cosine term to the disturbing function up to an optional max_order in the eccentricities and inclinations.

        Arguments
        ---------
        k_vec : Tuple of 6 ints 
            Coefficients (k1, k2, k3, k4, k5, k6) for a cosine term of the form 
            cos(k1*lambdaOut + k2*lambdaIn + k3*pomegaIn + k4*pomegaOut + k5*OmegaIn + k6*OmegaOut)
            Note the ks must sum to zero, and by convention we write lambdaOut first, e.g., (3, -2, -1, 0, 0, 0) is
            cos(3*lambdaOut - 2*lambdaIn - pomega1)
        max_order : int, optional
            Maximum order to go to in the (combined) eccentricities and inclinations.
            Can specify either max_order OR nu_vecs (see below), but not both (most users will use max_order).
            If neither are passed, max_order defaults to the leading order of the cosine term specified by k_vec.
        l_max : int, optional
            Maximum degree of expansion in :math:`\delta = (\Lambda-\Lambda_0)/\Lambda_0` to include in cosine coefficients. Default is 0.
            Can specify either l_max OR l_vecs (see below), but not both (most users will use l_max).
        nu_vecs: List of 4-tuples, optional
            A list of the specific combinations of nu indices to include for the cosine term coefficient, e.g., [(0, 0, 0, 0), (1, 0, 0, 0), ...]
            See paper and examples for definition and use.
            Can specify either max_order OR nu_vecs, but not both (max_order makes more sense for most use cases).
        l_vecs: List of 2-tuples, optional
            A list of the specific combinations of l indices to include for the cosine term coefficient, e.g., [(0, 0), (1, 0), (2, 0), ...] 
            See paper and examples for definition and use.
            Can specify either l_max OR l_vecs, but not both (l_max makes more sense for most use cases).
            One use case for passing particular l_vecs is if one body is massless, so the other Lambda doesn't vary (and doesn't need expansion)
        indexIn : int, optional
            Index of inner planet. Default 1.
        indexOut : int, optional
            Index of outer planet. Default 2.
        eccentricities: bool, optional
            By default, includes all eccentricity terms.
            Can set to False to exclude any eccentricity terms (e.g., fixed circular orbits).
        inclinations: bool, optional
            By default, includes all inclination terms.
            Can set to False to exclude any inclination terms (e.g., co-planar systems).
        """
        k1,k2,k3,k4,k5,k6 = k_vec
        k_vec = tuple(k_vec) # ensure k_vec is a tuple
        if np.sum(k_vec) != 0:
            raise AttributeError("Invalid k_vec={0}. The coefficients must sum to zero to satisfy the d'Alembert relation (rotational symmetry).".format(k_vec))
        if (k5+k6) % 2 != 0:
            raise AttributeError("Invalid k_vec={0}. Last two entries ({1} and {2}) must be divisible by 2 to maintain problem's reflection symmetry.".format(k_vec, k5, k6))
        if nu_vecs:
            nu_vecs = [tuple(nu_vec) for nu_vec in nu_vecs]
            if max_order:
                raise AttributeError('Must pass EITHER max_order OR nu_vecs to add_cos_term, but not both. See docstring.')
        else:
            # this is the leading order for the cosine term chosen (k3+k4+k5+k6 = -(k1+k2) by d'Alembert)
            min_order = abs(k3) + abs(k4) + abs(k5) + abs(k6)
            if not max_order:
                max_order = min_order
            if max_order < min_order:
                raise AttributeError("Passed a k_vec={0} whose order={1} is higher than the specified max_order={2}".format(k_vec, min_order, max_order))
            nu_max = (max_order - min_order)//2 # each nu contributes 2 powers of e or i, so divide by 2 rounding down
            nu_vecs = []
            for nu_tot in range(nu_max+1):
                nu_vecs += _nucombos(nutot=nu_tot) # Make list of [(nu1, nu2, nu3, nu4), ... ] tuples
        if l_vecs:
            l_vecs = [tuple(l_vec) for l_vec in l_vecs]
            if l_max > 0:
                raise AttributeError('Can only pass l_max OR l_vecs to add_cos_term. See docstring.')
        else:
            l_vecs = []
            for l_tot in range(l_max+1):
                l_vecs += _lcombos(ltot=l_tot) # Make list of [(l1, l2), ... ] tuples
      
        try:
            if np.array(l_vecs).shape[1] != 2:
                raise
        except:
            raise AttributeError("l_vecs = {0} must be a list of (l1, l2) pairs, e.g., [(0, 0), (0, 1), ...]".format(l_vecs))
        if np.array(nu_vecs).shape[1] != 4:
            raise AttributeError("nu_vecs = {0} must be a list of (nu1, nu2, nu3, nu4) tuples, e.g., [(0, 0, 0, 0), (1, 0, 0, 0), ...]".format(nu_vecs))
        lmax = np.max([l1+l2 for l1, l2 in l_vecs]) # maximum total l=l1+l2

        G = symbols('G')
        mIn,muIn,MIn,LambdaIn,lambdaIn,kappaIn,etaIn,sigmaIn,rhoIn = symbols('m{0},mu{0},M{0},Lambda{0},lambda{0},kappa{0},eta{0},sigma{0},rho{0}'.format(indexIn)) 
        mOut,muOut,MOut,LambdaOut,lambdaOut,kappaOut,etaOut,sigmaOut,rhoOut = symbols('m{0},mu{0},M{0},Lambda{0},lambda{0},kappa{0},eta{0},sigma{0},rho{0}'.format(indexOut)) 
        
        Lambda0In,Lambda0Out = _get_Lambda0_symbol(indexIn),_get_Lambda0_symbol(indexOut)
        alpha_sym = symbols(r"\alpha_{{{0}\,{1}}}".format(indexIn,indexOut))
        alpha_val = self.H_params[alpha_sym]
        aOut0 = _get_a0_symbol(indexOut)
        deltaIn = (LambdaIn - Lambda0In) / Lambda0In
        deltaOut = (LambdaOut - Lambda0Out) / Lambda0Out
        # alpha = aIn/aOut
        # Resonance components
        
        for nu_vec in nu_vecs:
            if eccentricities == False and k_nu_depend_on_eccentricities(k_vec, nu_vec) == True:
                continue
            if inclinations == False and k_nu_depend_on_inclinations(k_vec, nu_vec) == True:
                continue

            C_dict = df_coefficient_C(*k_vec, *nu_vec)#k1,k2,k3,k4,k5,k6,nu1,nu2,nu3,nu4)
            p1,p2 = _p1_p2_from_k_nu(k_vec,nu_vec)
            C_delta_expansion_dict = evaluate_df_coefficient_delta_expansion(C_dict,p1,p2,lmax,alpha_val)
            Ctot = 0
            for l_vec,C_val in C_delta_expansion_dict.items():
                if l_vec not in l_vecs: # have to calculate all terms up to lmax, but only consider if in l_vecs
                    continue
                if l_vecs and (indexIn,indexOut,(k_vec,nu_vec,l_vec)) in self.resonance_indices:
                    warnings.warn("Cosine term k_vec={0}, nu_vec={1}, l_vec={2} already included Hamiltonian; no new term added.".format(k_vec, nu_vec, l_vec))
                    continue
                if l_vecs and (indexIn,indexOut,(tuple([-k for k in k_vec]),nu_vec,l_vec)) in self.resonance_indices:
                    warnings.warn("The same cosine term k_vec={0}, nu_vec={1}, l_vec={2} (with all k_vec entries made negative) already included Hamiltonian; no new term added to avoid duplication (cos(x) = cos(-x)).".format(k_vec, nu_vec, l_vec))
                    continue
                self.resonance_indices.append((indexIn,indexOut,(k_vec,nu_vec,l_vec)))
                Csym = get_df_coefficient_symbol(*k_vec,*nu_vec,*l_vec,indexIn,indexOut)
                self.H_params[Csym] = C_val
                l1,l2=l_vec
                Ctot += Csym * deltaIn**l1 * deltaOut**l2
            nu1,nu2,nu3,nu4 = nu_vec
            rtLIn = sqrt(Lambda0In)
            rtLOut = sqrt(Lambda0Out)
            xin,yin = get_re_im_components(kappaIn/rtLIn ,-etaIn / rtLIn,k3)
            xout,yout = get_re_im_components( kappaOut/rtLOut, -etaOut/rtLOut,k4)
            uin,vin = get_re_im_components(sigmaIn/rtLIn/2, -rhoIn/rtLIn/2,k5)
            uout,vout = get_re_im_components(sigmaOut/rtLOut/2, -rhoOut/rtLOut/2,k6)

            re = uin*uout*xin*xout - vin*vout*xin*xout - uout*vin*xout*yin - uin*vout*xout*yin - uout*vin*xin*yout - uin*vout*xin*yout - uin*uout*yin*yout + vin*vout*yin*yout
            im = uout*vin*xin*xout + uin*vout*xin*xout + uin*uout*xout*yin - vin*vout*xout*yin + uin*uout*xin*yout - vin*vout*xin*yout - uout*vin*yin*yout - uin*vout*yin*yout
            
            GammaIn = (kappaIn*kappaIn + etaIn*etaIn)/2
            GammaOut = (kappaOut*kappaOut + etaOut*etaOut)/2
            QIn = (sigmaIn*sigmaIn + rhoIn*rhoIn)/2
            QOut = (sigmaOut*sigmaOut + rhoOut*rhoOut)/2
            
            eIn_sq_term = (2 * GammaIn / Lambda0In )**nu3
            eOut_sq_term = (2 * GammaOut / Lambda0Out )**nu4
            incIn_sq_term = ( QIn / Lambda0In / 2 )**nu1
            incOut_sq_term = ( QOut / Lambda0Out / 2 )**nu2
            
            # Update internal Hamiltonian
            prefactor1 = -G * mIn * mOut / aOut0
            prefactor2 = eIn_sq_term * eOut_sq_term * incIn_sq_term * incOut_sq_term 
            trig_term = re * cos(k1 * lambdaOut + k2 * lambdaIn) - im * sin(k1 * lambdaOut + k2 * lambdaIn) 
            
            self.H += prefactor1 * Ctot * prefactor2 * trig_term
    
    def add_MMR_terms(self,p,q,max_order=None,l_max=0,indexIn=1,indexOut=2,eccentricities=True,inclinations=True):
        """
        Add all eccentricity and inclination disturbing function terms associated with a p:p-q mean
        motion resonance up to a given order.

        Arguments
        ---------
        p : int
            Coefficient of lambdaOut in resonant argument
                j*lambdaOut - (j-k)*lambdaIn
        q : int
            Order of the mean motion resonance.
        max_order : int
            Maximum order of terms to add.
        l_max : int, optional
            Maximum degree of expansion in :math:`\delta = (\Lambda-\Lambda_0)/\Lambda_0` to include in cosine coefficients. Default is 0.
        indexIn : int
            Index of inner planet.
        indexOut : int
            Index of outer planet.
        eccentricities: bool, optional
            By default, includes all eccentricity terms.
            Can set to False to exclude any eccentricity terms (e.g., fixed circular orbits).
        inclinations: bool, optional
            By default, includes all inclination terms.
            Can set to False to exclude any inclination terms (e.g., co-planar systems).
        """
        if p<q or q<0:
            warnings.warn("""
            MMRs with p<q or q<0 are not supported. 
            If you really want to include these terms, 
            they may be added individually with the 
            'add_cosine_term' method.
            """)
        if abs(p) % q == 0 and q != 1:
            warnings.warn("p and q share a common divisor. Some important terms may be omitted!")
        
        if not max_order:
            max_order = q
        if max_order < q:
            warnings.warn("""Maximum order is lower than order of the resonance!""")
        assert max_order>0, "max_order= {:d} not allowed, must be non-negative.".format(max_order)
  
        terms = list_resonance_terms(p,q,min_order=q,max_order=max_order,eccentricities=eccentricities,inclinations=inclinations)
        for k_vec, nu_vec in terms:
            self.add_cosine_term(k_vec=k_vec,l_max=l_max,nu_vecs=[nu_vec],indexIn=indexIn,indexOut=indexOut)
    
    def add_secular_terms(self,min_order=2,max_order=2,l_max=0,indexIn=1,indexOut=2,eccentricities=True,inclinations=True):
        """
        Add all eccentricity and inclination secular terms in the disturbing function from
        min_order to max_order in the eccentricities and inclinations.

        Arguments
        ---------
        min_order : int, optional
            Minimum order terms in the eccentricities and inclinations to add. Defaults to 2. 
        max_order : int, optional
            Maximum order terms in the eccentricities and inclinations to add. Defaults to 2.
        l_max : int, optional
            Maximum degree of expansion in :math:`\delta = (\Lambda-\Lambda_0)/\Lambda_0` to include in cosine coefficients. Default is 0.
            to include in cosine coefficients. Default is 0.
        indexIn : int
            Index of inner planet.
        indexOut : int
            Index of outer planet.
        eccentricities: bool, optional
            By default, includes all eccentricity terms.
            Can set to False to exclude any eccentricity terms (e.g., fixed circular orbits).
        inclinations: bool, optional
            By default, includes all inclination terms.
            Can set to False to exclude any inclination terms (e.g., co-planar systems).
        """
        assert max_order>0, "max_order= {:d} not allowed,  must be non-negative.".format(max_order)
  
        terms = list_secular_terms(min_order=min_order,max_order=max_order,eccentricities=eccentricities,inclinations=inclinations)
        for k_vec, nu_vec in terms:
            self.add_cosine_term(k_vec=k_vec,l_max=l_max,nu_vecs=[nu_vec],indexIn=indexIn,indexOut=indexOut)

    def add_orbit_average_J2_terms(self,J2,Rin,max_ei_order=None,max_delta_order=None,particles = 'all',**kwargs):
        r"""
        Add Hamiltonian terms that capture the orbit-averaged effect of 
        a central body's oblateness parameterized by the :math:`J_2`
        gravitational harmonic.

        Arguments
        ---------
        J2 : float
            The value of the central body's J2 gravitational
            harmonic.
        Rin : float
            The central body's radius
        max_ei_order : int, optional
            Maximum order of expansion in eccentricity and inclination.
            By default, the value is set to 'None' and no expansion in 
            eccentricity and inclinaion is done.
        max_delta_order : int, optional
            Maxmimum order in :math:`\delta =(\Lambda-\Lambda_0)/\Lambda_0).
            Default is 'None' and dependence on :math:`Lambda` is exact.
        particles : list, optional
            Which particle numbers to add :math:`J_2` terms for. Default
            is set to all particles.
        """
        G = symbols('G')
        J2_s = kwargs.get("J2_symbol",symbols("J2"))
        Rin_s = kwargs.get("Rin_symbol",symbols(r"R"))
        self.H_params[J2_s] = J2
        self.H_params[Rin_s] = Rin
        GJ2RinSq = G * J2_s * Rin_s * Rin_s 
        a0_d = symbols("a0")
        # dummy variables, substitute later
        # Lambda_d = symbols("Lambda")
        Lambda0_d = symbols("Lambda0")
        delta_d = symbols("delta")
        Lambda_d = Lambda0_d * (1 + delta_d)
        kappa_d,eta_d,sigma_d,rho_d = symbols('kappa,eta,sigma,rho')
        Gamma = (kappa_d * kappa_d + eta_d * eta_d) / 2
        G = Lambda_d - Gamma
        omesq = (G / Lambda_d) * (G / Lambda_d)
        esq = 1 - omesq
        Q = (sigma_d*sigma_d + rho_d*rho_d) / 2
        cosI = 1 - Q / G
        ssq = (1 - cosI) /2 
        a = a0_d * (1 + delta_d) * (1 + delta_d)
        rt_omesq = sqrt(1 - esq).expand()
        num = 3 * (ssq*ssq - ssq) + 1/S(2)
        denom = rt_omesq * rt_omesq * rt_omesq 
        full_exprn =  num / denom / a / a / a
        # Expand to max order if specified.
        # Otherwise the complete expression is used.
        if max_ei_order:
            eps = symbols("epsilon")
            # e/i expansion
            eps_exprn = full_exprn.subs({sym:eps*sym for sym in (kappa_d,eta_d,sigma_d,rho_d)})
            full_exprn = series(eps_exprn,eps,0,max_ei_order+1).removeO().subs({eps:1})
        if max_delta_order:
            # delta expansion
            full_exprn = series(full_exprn,delta_d,0,max_delta_order+1).removeO()
        Hpert = -1 * GJ2RinSq * full_exprn 
        if particles == 'all':
            pids = range(1,self.N)
        else:
            pids = particles
        for pid in pids:
            p = self.particles[pid]
            m,mu,M,kappa,eta,sigma,rho = symbols('m{0},mu{0},M{0},kappa{0},eta{0},sigma{0},rho{0}'.format(pid)) 
            Lambda = symbols('Lambda{0}'.format(pid)) 
            Lambda0 = _get_Lambda0_symbol(pid)
            delta = (Lambda - Lambda0)/Lambda0
            a0 = _get_a0_symbol(pid)
            delta = (Lambda - Lambda0)/Lambda0
            self.H += M * mu * Hpert.subs({a0_d:a0,delta_d:delta,kappa_d:kappa,eta_d:eta,sigma_d:sigma,rho_d:rho,Lambda0_d:Lambda0})

    def add_gr_potential_terms(self,c,max_e_order=None,particles = 'all'):
        r"""
        Add Hamiltonian terms that capture the orbital precession
        caused by general relativity by adding a potential term
        of the form 
        
        .. math::
             \phi_\mathrm{GR} = \frac{3G^2M_*^2}{c^2a^2\sqrt{1-e^2}}

        Arguments
        ---------
        c : float
            The speed of light in the appropriate simulation units.
            harmonic.
        max_e_order : int, optional
            Maximum order of expansion in eccentricity.
            By default, the value is set to 'None' and no expansion in 
            eccentricity is done.
        particles : list, optional
            Which particle numbers to add :math:`J_2` terms for. Default
            is set to all particles.
        """
        G,c_s = symbols('G,c')
        G_by_c = G / c_s
        self.H_params[c_s] = c
        
        # dummy variables, substitute later
        # Lambda_d = symbols("Lambda")
        a0_d = symbols("a0")
        Lambda0_d = symbols("Lambda0")
        kappa_d,eta_d,sigma_d,rho_d = symbols('kappa,eta,sigma,rho')
        Gamma = (kappa_d * kappa_d + eta_d * eta_d) / 2
        G = Lambda0_d - Gamma
        omesq = (G / Lambda0_d) * (G / Lambda0_d)
        esq = 1 - omesq
        full_exprn =  -3 * G_by_c * G_by_c / a0_d / a0_d / sqrt(omesq)
        # Expand to max order if specified.
        # Otherwise the complete expression is used.
        if max_e_order:
            eps = symbols("epsilon")
            # e expansion
            eps_exprn = full_exprn.subs({sym:eps*sym for sym in (kappa_d,eta_d)})
            full_exprn = series(eps_exprn,eps,0,max_e_order+1).removeO().subs({eps:1})
        if particles == 'all':
            pids = range(1,self.N)
        for pid in pids:
            p = self.particles[pid]
            m,mu,M,kappa,eta = symbols('m{0},mu{0},M{0},kappa{0},eta{0}'.format(pid)) 
            Lambda0 = _get_Lambda0_symbol(pid)
            a0 = _get_a0_symbol(pid)
            self.H += M * M * mu * full_exprn.subs({a0_d:a0,kappa_d:kappa,eta_d:eta,Lambda0_d:Lambda0})

    def _get_laplace_lagrange_matrices(self):
        set_e_and_inc_zero_rule = {
            S('{0}{1}'.format(var,i)):0
           for i in range(1,self.N)
            for var in ['eta','kappa','rho','sigma']
        }
        mtrx = []
        for s1 in [S('eta{}'.format(i)) for i in range(1,self.N)]:
            row = []
            for s2 in [S('kappa{}'.format(i)) for i in range(1,self.N)]:
                entry= diff(self.derivs[s1],s2)
                row.append(entry.subs(set_e_and_inc_zero_rule))
            mtrx.append(row)
        ecc_mtrx = Matrix(mtrx)
        mtrx = []
        for s1 in [S('rho{}'.format(i)) for i in range(1,self.N)]:
            row = []
            for s2 in [S('sigma{}'.format(i)) for i in range(1,self.N)]:
                entry= diff(self.derivs[s1],s2)
                row.append(entry.subs(set_e_and_inc_zero_rule))
            mtrx.append(row)
        inc_mtrx = Matrix(mtrx)
        return ecc_mtrx,inc_mtrx



from warnings import warn
import numpy as np
from .miscellaneous import poisson_bracket
from sympy import lambdify, solve, diff,sin,cos,sqrt,atan2,symbols,Matrix, Mul,S
from sympy import trigsimp, simplify, postorder_traversal, powsimp, factor_terms
from sympy.simplify.fu import TR10,TR10i
from .hamiltonian import reduce_hamiltonian, PhaseSpaceState, Hamiltonian
from .poincare import Poincare
from sympy import Wild,atan2,exp,sin,cos,Function,sqrt,expand,simplify,Add
def _cos_n_atan(x,y,n):
    """
    Express cos(n*atan2(y,x)) in terms
    of x and y without trig functions.
    """
    assert n.is_integer
    if n<0:
        return _cos_n_atan(x,y,-1*n)
    if n==0:
        return 1
    else:
        r2 = x*x+y*y
        r = sqrt(r2)
        return x * _cos_n_atan(x,y,n-1)/r - y * _sin_n_atan(x,y,n-1)/r
def _sin_n_atan(x,y,n):
    """
    Express sin(n*atan2(y,x)) in terms
    of x and y without trig functions.
    """
    assert n.is_integer
    if n<0:
        return -1 * _sin_n_atan(x,y,-1*n)
    if n==0:
        return 0
    else:
        r2 = x*x+y*y
        r = sqrt(r2)
        return x*_sin_n_atan(x,y,n-1)/r + y*_cos_n_atan(x,y,n-1)/r

def _simplify_atans(exprn):
    """
    Expand any occurences of cos(a + n atan(y,x)) 
    and sin(a + n atan(y,x)).
    """
    if not _exprn_contains_funcs(exprn,[atan2]):
        return factor_terms(exprn) 
    pnames = ('a','x','y')
    a_w,x_w,y_w = symbols(','.join(n_+"_w" for n_ in pnames),cls=Wild,real=True)
    n_w = Wild('n_w',properties=(lambda x: x.is_integer,))
    c = Function("c",real=True)
    s = Function("s",real=True)
    cos_patt1 = cos(n_w * atan2(y_w,x_w))
    sin_patt1 = sin(n_w * atan2(y_w,x_w))
    res = TR10(exprn)
    res = res.replace(cos_patt1,c(x_w,y_w,n_w))
    res = res.replace(sin_patt1,s(x_w,y_w,n_w))
    res = res.replace(c,_cos_n_atan)
    res = res.replace(s,_sin_n_atan)
    res = factor_terms(TR10i(res))
    return res

def _exprn_contains_funcs(exprn,funcs):
    for arg in postorder_traversal(exprn):
        if arg.func in funcs:
            return True
    return False

def _trigsimp(_):
    if _exprn_contains_funcs(_,[sin,cos]):
        _ = expand(_.rewrite(exp))
        _ = powsimp(_)
        _ = factor_terms(_)
        _ =_.rewrite(cos)
    return _


def _termwise_trigsimp(exprn):
    if not hasattr(exprn,'args'):
        return exprn
    if len(exprn.args)==0:
        return exprn
    return exprn.func(*[_trigsimp(a) for a in exprn.args])

def _termwise_atan2simp(exprn):
    if not hasattr(exprn,'args'):
        return exprn
    if len(exprn.args)==0:
        return exprn
    return exprn.func(*[_simplify_atans(a) for a in exprn.args])

def _get_default_qpxy_symbols(N,cartesian_indices):
    qppairs = _get_default_qp_symbols(N)
    xypairs = _get_default_xy_symbols(N)
    func = lambda i: (xypairs[i][1],xypairs[i][0]) if i in cartesian_indices else qppairs[i]
    varpairs = list(map(func,range(N)))
    return varpairs
def _get_default_qp_symbols(N):
    default_qp_symbols=list(zip(
        symbols("Q(1:{})".format(N+1),real=True),
        symbols("P(1:{})".format(N+1),real=True)
    ))
    return default_qp_symbols

def _get_default_xy_symbols(N):
    default_xy_symbols=list(zip(
        symbols("x(1:{})".format(N+1),real=True),
        symbols("y(1:{})".format(N+1),real=True)
    ))
    return default_xy_symbols

class CanonicalTransformation():
    r"""
    A class for representing canonical transformations.
    
    An instance represents a transformation from old canonical variables,
    :math:`(q,p)` to new variables :math:`(q',p')`.

    Attributes
    ----------
    old_qp_vars : list
        A list of symbols representing the old variables, :math:`(q,p)`.
        The list should contain all coordinate variables, :math:`q`, followed
        by their conjugate momentum variables :math:`p`.
    new_qp_vars : list
        A list of symbols representing the new variables, :math:`(q',p')`.
        The list should contain all coordinate variables, :math:`q'`, followed
        by their conjugate momentum variables :math:`p'`.
    old_to_new_rule : dictionary
        A dictionary defining the substitution rules applied to transform from
        old canonical variables to new ones.
    new_to_old_rule : dictionary
        A dictionary defining the substitution rules applied to transform from
        new canonical variables to old ones.
    old_to_new_simplify : function
        A function that will be applied to expressions when transforming from
        old to new variables.
    new_to_old_simplify : function
        A function that will be applied to expression when transforming from
        new to old variables.
    params : dict
        A dictionary setting the numerical values of any parameters on which
        the transformation depends.
    H_scale : symbol or float
        A factor by which the transformation rescales the Hamiltonian. This is
        used when applying transformations that simulatneously rescale the
        momenta and the Hamiltonian.
    """
    def __init__(self,
            old_qp_vars,
            new_qp_vars,
            old_to_new_rule = None,
            new_to_old_rule = None,
            old_to_new_simplify = None,
            new_to_old_simplify = None,
            params = {},
            **kwargs
    ):
        self.old_qp_vars = old_qp_vars
        self.new_qp_vars = new_qp_vars
        self.params = params
        self.H_scale = kwargs.get('H_scale',1)

        if not (old_to_new_rule or new_to_old_rule):
            raise ValueError("Must specify at least one of 'old_to_new_rule' or 'new_to_old_rule'!")
        if old_to_new_rule:
            self.old_to_new_rule = old_to_new_rule
        else:
            self.new_to_old_rule = new_to_old_rule
            eqs = [v - self.new_to_old_rule[v] for v in self.new_qp_vars]
            self.old_to_new_rule = solve(eqs,self.old_qp_vars)
        
        if new_to_old_rule:
            self.new_to_old_rule = new_to_old_rule
        else:
            eqs = [v - self.old_to_new_rule[v] for v in self.old_qp_vars] 
            self.new_to_old_rule = solve(eqs,self.new_qp_vars)

        if old_to_new_simplify:
            self.old_to_new_simplify_function = old_to_new_simplify
        else:
            self.old_to_new_simplify_function = lambda x: x

        if new_to_old_simplify:
            self.new_to_old_simplify_function = new_to_old_simplify
        else:
            self.new_to_old_simplify_function = lambda x: x

        qpv_new = self.new_qp_vars
        qpv_old = self.old_qp_vars
        self._old_to_new_vfunc = lambdify(qpv_old,[self.N_new_to_old(v) for v in qpv_new])
        self._new_to_old_vfunc = lambdify(qpv_new,[self.N_old_to_new(v) for v in qpv_old])
   
    def old_to_new(self,exprn):
        r"""
        Transform an expression in terms of old canonical variables to one in
        terms of new varialbes.

        Arguments
        ---------
        exprn : sympy expression
            The expression to transform
        Returns
        -------
        sympy expression
            The transformed expression.
        """
        exprn = exprn.subs(self.old_to_new_rule)
        exprn = self.old_to_new_simplify_function(exprn)
        return exprn

    def N_old_to_new(self,exprn):
        r"""
        Same as
        :func:`~celmech.canonical_transformations.CanonicalTransformation.old_to_new`
        but with numerical values replacing any symbolic parameters.
        """
        exprn = self.old_to_new(exprn)
        Nexprn = exprn.subs(self.params)
        return Nexprn

    def new_to_old(self,exprn):
        r"""
        Transform an expression in terms of new canonical variables to one in
        terms of old varialbes.

        Arguments
        ---------
        exprn : sympy expression
            The expression to transform.
        Returns
        -------
        sympy expression
            The transformed expression.
        """
        exprn = exprn.subs(self.new_to_old_rule)
        exprn = self.new_to_old_simplify_function(exprn)
        return exprn
    
    def N_new_to_old(self,exprn):
        r"""
        Same as
        :func:`~celmech.canonical_transformations.CanonicalTransformation.new_to_old`
        but with numerical values replacing any symbolic parameters.
        """
        exprn = self.new_to_old(exprn)
        Nexprn = exprn.subs(self.params)
        return Nexprn

    def old_to_new_array(self,arr):
        r"""
        Convert an array of old variable values to the corresponding values of
        new canonical variables.

        Arguments
        ---------
        arr : array-like
            Value of old variables

        Returns
        -------
        new_arr : numpy.array
            Array of new variable values.
        """
        return np.array(self._old_to_new_vfunc(*arr))

    def new_to_old_array(self,arr):
        r"""
        Convert an array of new variable values to the corresponding values of
        old canonical variables.

        Arguments
        ---------
        arr : array-like
            Value of new variables

        Returns
        -------
        new_arr : numpy.array
            Array of old variable values.
        """
        return np.array(self._new_to_old_vfunc(*arr))

    def new_to_old_state(self,state):
        old_values = self.new_to_old_array(state.values)
        old_state = PhaseSpaceState(self.old_qp_vars,old_values,t=state.t)
        return old_state
    
    def old_to_new_hamiltonian(self,ham,do_reduction = False):
        r"""
        Apply canonical transformation to a
        :class:`~celmech.hamiltonian.Hamiltonian` object, transforming the
        Hamiltonian and producing a new object with a
        :attr:`~celmech.hamiltonian.Hamiltonian.state` attribute representing
        the phase space position in the new canonical variables.

        Arguments
        ---------
        ham : celmech.hamiltonian.Hamiltonian
            The Hamiltonian object to which transformation is applied.
        do_reduction : bool, optional
            If ``True``, automatically reduce the number of degrees of freedom
            by detecting which canonical variables do not appear in the
            expression of the transformed Hamiltonian.

        Returns
        -------
        new_ham : celmech.hamiltonian.Hamiltonian
            The transformed Hamiltonian.
        """
        # Get full values
        new_values = self.old_to_new_array(ham.full_values)
        # The full set of q,p variables
        old_full = set(ham.full_qp_vars)
        # Only q,p variables that are active
        old_active = set(ham.qp_vars)
        # Ignorable q,p variables
        old_ignorable = old_full.difference(old_active)

        # New, transformed Hamiltonian expression
        newH =  self.old_to_new(ham.H)
        
        new_pars = ham.H_params.copy()
        new_pars.update(self.params)

        # Handle possible cases where ignorable coordinates get transformed:
        #   First, remove old ingorable coordinates from the new H_params dictionary
        for ig in old_ignorable:
            new_pars.pop(ig,None)
        #   Next, loop through new variable pairs and detect ingorable ones.
        #       'is_active' stores a boolean array indicating which new q,p vars 
        #       will be active variables.
        is_active = np.ones(2*self.N_dof,dtype=bool)
        for i,qpnew in enumerate(self.new_qp_pairs):
            qnew,pnew = qpnew
            vars_set_qnew = self.new_to_old(qnew).free_symbols
            vars_set_pnew = self.new_to_old(pnew).free_symbols
            vars_set_qpnew = vars_set_qnew.union(vars_set_pnew)
            # A new DOF is ignorable if its expression in terms of 
            # old variables does not involve any of the old active 
            # variables.
            is_ignorable = len(old_active.intersection(vars_set_qpnew)) == 0 
            if is_ignorable:
                is_active[i] = False
                is_active[i+self.N_dof] = False
                #if qnew in newH.free_symbols:
                new_pars[qnew] = new_values[i]
                #if pnew in newH.free_symbols:
                new_pars[pnew] = new_values[i+self.N_dof]

        # The list of new active q,p variables.
        new_vars_reduced = [var for var,activeQ in zip(self.new_qp_vars,is_active) if activeQ]

        # New state consisting only of active variables.
        new_state = PhaseSpaceState(new_vars_reduced,new_values[is_active],ham.state.t)
        

        # Rescale Hamiltonian.
        # Warning--- this will not work properly if the top-level function is not sympy.Add
        # This should probably be treated more carefully.
        if self.H_scale != 1:
            assert newH.func == Add,"Re-scaling requires the top level node of the Hamiltonian expression tree must be sympy.Add"
        newH = newH.func(*[a*self.H_scale for a in newH.args])

        # New Hamiltonian: the new state only contains active variables
        # but full_qp_vars kwarg is passed the full set of new q,p variables.
        new_ham = Hamiltonian(newH,new_pars,new_state,full_qp_vars = self.new_qp_vars)
        if do_reduction:
            new_ham = reduce_hamiltonian(new_ham)
        return new_ham
    
    def new_to_old_hamiltonian(self,ham,do_reduction = False):
        old_state = self.new_to_old_state(ham.state)
        oldH = self.new_to_old(ham.H) / self.H_scale
        old_ham = Hamiltonian(oldH,ham.H_params,old_state)
        if do_reduction:
            old_ham = reduce_hamiltonian(old_ham)
        return old_ham
    
    def _test_new_to_old_canonical(self):
        pb = lambda q,p: poisson_bracket(q,p,self.old_qp_vars,[]) / self.H_scale
        bracket_results = [pb(self.N_new_to_old(qq),self.N_new_to_old(pp)).simplify() for qq,pp in self.new_qp_pairs]
        br_arr = [np.isclose(float(x.subs(self.params)),1,atol=1e-10) for x in bracket_results]
        return np.alltrue(br_arr)

    def _test_old_to_new_canonical(self):
        pb = lambda q,p: poisson_bracket(q,p,self.new_qp_vars,[]) * self.H_scale
        bracket_results = [pb(self.old_to_new(qq),self.old_to_new(pp)).simplify() for qq,pp in self.old_qp_pairs]
        br_arr = [np.isclose(float(x.subs(self.params)),1,atol=1e-10) for x in bracket_results]
        return np.alltrue(br_arr)
    def test_canonical(self):
        """
        Test whether the substitution rules of this tranformation constitute 
        a canonical transformation.

        Returns
        -------
        bool :
            Returns ``True`` if the transformation is canonical.
        """
        return self._test_new_to_old_canonical() and self._test_old_to_new_canonical()
    @property 
    def N_dof(self):
        return int(len(self.old_qp_vars)/2)

    @property
    def old_qp_pairs(self):
        return [(self.old_qp_vars[i], self.old_qp_vars[i+self.N_dof]) for i in range(self.N_dof)]
    
    @property
    def new_qp_pairs(self):
        return [(self.new_qp_vars[i], self.new_qp_vars[i+self.N_dof]) for i in range(self.N_dof)]

    @classmethod
    def from_type2_generating_function(cls,F2func,old_qp_vars,new_qp_vars,**kwargs):
        r"""
        Initialize a canonical transformation derived from a `type 2 generating
        function`_. Given a set of old variables :math:`(q,p)`, new variables,
        :math:`(Q,P)` and generating function :math:`F_2(q,P)`, the
        transformation rules are given by

        .. math::
            p &=& \frac{\partial }{\partial q}F_2(q,P) \\ 
            Q &=& \frac{\partial }{\partial P}F_2(q,P) 

        .. _type 2 generating function: https://en.wikipedia.org/wiki/Canonical_transformation#Type_2_generating_function

        Arguments
        ---------
        F2func : sympy expression
            The type 2 generating function
        old_qp_vars : array-like
            The list of old canonical variables
        new_qp_vars : array-like
            The list of new canonical variables

        Returns
        -------
        .CanonicalTransformation
            The resulting transformation.
        """
        N_dof = int(len(old_qp_vars)/2)
        old_qp_pairs = [(old_qp_vars[i], old_qp_vars[i+N_dof]) for i in range(N_dof)]
        new_qp_pairs = [(new_qp_vars[i], new_qp_vars[i+N_dof]) for i in range(N_dof)]
        eqs = [p - diff(F2func,q) for q,p in old_qp_pairs]
        eqs += [Q - diff(F2func,P) for Q,P in new_qp_pairs]
        o2n_soln = solve(eqs,old_qp_vars)

        if type(o2n_soln) is dict:
            o2n = o2n_soln
        elif type(o2n_soln) is list:
            msg ="Multiple roots found when solving for old variables in"
            msg+="of new. The desired transformation may not be produced"
            warn(msg)
            o2n = dict(zip(old_qp_vars,o2n_soln[0]))

        n2o_soln = solve(eqs,new_qp_vars)
        if type(n2o_soln) is dict:
            n2o = n2o_soln
        elif type(n2o_soln) is list:
            msg ="Multiple roots found when solving for new variables in"
            msg+="terms of old. The desired transformation may not be produced"
            warn(msg)
            n2o = dict(zip(new_qp_vars,n2o_soln[0]))

        return cls(old_qp_vars,new_qp_vars,old_to_new_rule=o2n,new_to_old_rule = n2o)
    
    @classmethod
    def cartesian_to_polar(cls,old_qp_vars,indices=None,polar_symbol_pairs=None,**kwargs):
        r"""
        Initialize a canonical transformation that transforms a user-specified
        subset of conjugate coordinate-momentum pairs, :math:`(q_i,p_i)`, to
        new 'polar-style' variable pairs

        .. math::
            (Q_i,P_i) = \left(\tan^{-1}(q_i/p_i), \frac{1}{2}(q_i^2+p_i^2)\right)

        Arguments
        ---------
        old_qp_vars : array-like
            The list of old canonical variables
        indices : array-list
            List of integer indices of the coordinate-momentum pairs to be
            transformed.
        polar_symbol_pairs : list, optional
            List of symbols to use for the new polar-style variables.
            Default symbols are :math:`(Q_i,P_i)` where the index i
            ranges over the number of variable pairs being transformed.

        Returns
        -------
        .CanonicalTransformation
            The resulting transformation.
        """
        N_dof = int(len(old_qp_vars)/2)
        old_qp_pairs = [(old_qp_vars[i], old_qp_vars[i+N_dof]) for i in range(N_dof)]
        if not indices:
            indices = range(N_dof)
        N = len(indices)
        if not polar_symbol_pairs:
            polar_symbol_pairs = _get_default_qp_symbols(N)
        new_q, new_p = [], []
        o2n,n2o = dict(),dict()
        for i,qp in enumerate(old_qp_pairs):
            if i in indices:
                y,x = qp 
                Q,P = polar_symbol_pairs.pop(0)
                o2n.update({ x:sqrt(2*P) * cos(Q) , y:sqrt(2*P) * sin(Q)})
                n2o.update({P:(x*x +y*y)/2, Q:atan2(y,x)})
                new_q.append(Q)
                new_p.append(P)
            # keep indentity transformation if index not passed
            else:
                id_tr = dict(zip(qp,qp))
                o2n.update(id_tr)
                n2o.update(id_tr)
                new_q.append(qp[0])
                new_p.append(qp[1])

        # add a simplify function to remove arctans
        atan2_simplify = lambda exprn: TR10i(simplify(TR10(exprn))) if _exprn_contains_funcs(exprn,[atan2]) else powsimp(exprn)
        n2o_simplify = lambda x: x.func(*[atan2_simplify(a) for a in x.args]) if len(x.args) > 0 else x
        kwargs.setdefault("old_to_new_simplify",_termwise_trigsimp)
        kwargs.setdefault("new_to_old_simplify",n2o_simplify)
        new_qp_vars = new_q + new_p
        return cls(old_qp_vars,new_qp_vars,o2n,n2o,**kwargs)
    
    @classmethod
    def polar_to_cartesian(cls,old_qp_vars,indices=None,cartesian_symbol_pairs=None,**kwargs):
        r"""
        Initialize a canonical transformation that transforms a user-specified
        subset of conjugate coordinate-momentum pairs, :math:`(q_i,p_i)`, to
        new 'cartesian-style' variable pairs

        .. math::
            (Q_i,P_i) = \left(\sqrt{2p_i}\sin q_i,\sqrt{2p_i}\cos q_i\right)

        Arguments
        ---------
        old_qp_vars : array-like
            The list of old canonical variables
        indices : array-list
            List of integer indices of the coordinate-momentum pairs to be
            transformed.
        cartesian_symbol_pairs : list, optional
            List of symbols to use for the new cartesian-style variables.
            Default symbols are :math:`(y_i,x_i)` where the index i ranges
            over the number of variable pairs being transformed.

        Returns
        -------
        .CanonicalTransformation
            The resulting transformation.
        """
        N_dof = int(len(old_qp_vars)/2)
        old_qp_pairs = [(old_qp_vars[i], old_qp_vars[i+N_dof]) for i in range(N_dof)]
        if not indices:
            indices = range(N_dof)
        N = len(indices)
        if not cartesian_symbol_pairs:
           cartesian_symbol_pairs = _get_default_xy_symbols(N)
        new_q, new_p = [], []
        o2n,n2o = dict(),dict()
        for i,qp in enumerate(old_qp_pairs):
            if i in indices:
                x,y = cartesian_symbol_pairs.pop(0)
                Q,P = qp 
                n2o.update({ x:sqrt(2*P) * cos(Q) , y:sqrt(2*P) * sin(Q)})
                o2n.update({P:(x*x +y*y)/2, Q:atan2(y,x)})
                new_q.append(y)
                new_p.append(x)
            # keep identity transformation if index not passed
            else:
                id_tr = dict(zip(qp,qp))
                o2n.update(id_tr)
                n2o.update(id_tr)
                new_q.append(qp[0])
                new_p.append(qp[1])

        # A simplify function to remove arctans
        o2n_simplify = _termwise_atan2simp
        #atan2_simplify = lambda exprn: TR10i(simplify(TR10(exprn))) if _exprn_contains_funcs(exprn,[atan2]) else powsimp(exprn)
        #o2n_simplify = lambda x: x.func(*[atan2_simplify(a) for a in x.args]) if len(x.args) > 0 else x
        kwargs.setdefault("old_to_new_simplify",o2n_simplify)
        new_qp_vars = new_q + new_p
        return cls(old_qp_vars,new_qp_vars,o2n,n2o,**kwargs)

    @classmethod
    def from_linear_angle_transformation(cls,old_qp_vars,Tmtrx,old_cartesian_indices=[],new_cartesian_indices=[],**kwargs):
        r"""
        Produces the transformation :math:`(\pmb{q},\pmb{p})\rightarrow(\pmb{Q},\pmb{P})` given by

        .. math::
            \begin{eqnarray}
                 \pmb{Q} = T \cdot \pmb{q} ~&;~
                 \pmb{P} = (T^{-1})^\mathrm{T} \cdot \pmb{p}
            \end{eqnarray}

        where :math:`T` is a user-specified matrix.


        Arguments
        ---------
        old_qp_vars : list
            List of old variable symbols.
        Tmtrx : array-like
            Matrix defining the canonical transformation.
        old_cartesian_indicies : list
            List of indices specifying which old variables are cartesian-style
            pairs.
        new_cartesian_indicies : list
            List of indices specify which new variables should be created as
            cartesian style pairs.

        Returns
        -------
        .CanonicalTransformation
        """
        try:
            N_dof = len(old_qp_vars)//2
            Tmtrx = np.array(Tmtrx).reshape(N_dof,N_dof)
        except:
            raise ValueError("'Tmtrx' could not be shaped into a {0}x{0} array".format(N_dof))
        old_qp_pairs = list(zip(old_qp_vars[:N_dof],old_qp_vars[N_dof:]))
        new_qp_pairs = kwargs.get("QPvars",_get_default_qpxy_symbols(N_dof,new_cartesian_indices))
        new_coords = Matrix([q for q,p in new_qp_pairs])
        new_momenta = Matrix([p for q,p in new_qp_pairs])

        to_angle = lambda i,qp,cartesian_indices: atan2(*qp) if i in cartesian_indices else qp[0] 
        to_action = lambda i,qp,cartesian_indices: (qp[0]**2 + qp[1]**2)/S(2) if i in cartesian_indices else qp[1] 

        old_angvars = Matrix([to_angle(i,qp,old_cartesian_indices) for i,qp in enumerate(old_qp_pairs)])
        old_actvars = Matrix([to_action(i,qp,old_cartesian_indices) for i,qp in enumerate(old_qp_pairs)])
        new_angvars = Matrix([to_angle(i,qp,new_cartesian_indices) for i,qp in enumerate(new_qp_pairs)])
        new_actvars = Matrix([to_action(i,qp,new_cartesian_indices) for i,qp in enumerate(new_qp_pairs)])

        Tmtrx = Matrix(Tmtrx)
        Tmtrx_inv  = Tmtrx.inv()

        n2o = dict()
        for i,qp_new,ang_exprn,act_exprn in zip(range(N_dof),new_qp_pairs,Tmtrx * old_angvars,Tmtrx_inv.T * old_actvars):
            qnew,pnew = qp_new
            if i in new_cartesian_indices:
                n2o[qnew] = simplify(sqrt(2*act_exprn) * TR10i(TR10(sin(ang_exprn))))
                n2o[pnew] = simplify(sqrt(2*act_exprn) * TR10i(TR10(cos(ang_exprn))))
            else:
                n2o[qnew] = ang_exprn
                n2o[pnew] = act_exprn
        
        o2n = dict()
        for i,qp_old,ang_exprn,act_exprn in zip(range(N_dof),old_qp_pairs,Tmtrx_inv * new_angvars,Tmtrx.T * new_actvars):
            qold,pold = qp_old
            if i in old_cartesian_indices:
                o2n[qold] = simplify(sqrt(2*act_exprn) * TR10i(TR10(sin(ang_exprn))))
                o2n[pold] = simplify(sqrt(2*act_exprn) * TR10i(TR10(cos(ang_exprn))))
            else:
                o2n[qold] = ang_exprn
                o2n[pold] = act_exprn
        
        new_ang_vars = set([new_qp_pairs[i][0] for i in range(N_dof) if i not in new_cartesian_indices])
        simplify_trig = lambda a: simplify(TR10i(a.expand())) if len(new_ang_vars.intersection(a.free_symbols))>0 else a
        o2n_simplify = lambda x: x.func(*[simplify_trig(arg) for arg in x.args]) if len(x.args)>0 else x
        kwargs.setdefault("old_to_new_simplify",o2n_simplify)
        return cls(old_qp_vars,list(new_coords) + list(new_momenta), o2n, n2o,**kwargs)

    @classmethod
    def from_poincare_angles_matrix(cls,p_vars,Tmtrx,new_qp_pairs=None,**kwargs):
        r"""
        Given an input :class:`~celmech.poincare.Poincare` instanace and an
        invertible matrix, :math:`T`, generate a transformation for which 
        the new angle variables are given by
        
        .. math::
            \pmb{Q} = T \cdot 
            \begin{pmatrix}
            \lambda_1 \\ \vdots \\ \lambda_N \\
            \gamma_1 \\ \vdots \\ \gamma_N \\
            q_1 \\ \vdots \\ q_N
            \end{pmatrix}~.

        Arguments
        ---------
        pvars : celmech.poincare.Poincare
            A Poincare instance with variables to which the resulting
            transformation will apply.
        Tmtrx : ndarray, shape (N,N)
            An invertible matrix determining the new canonical angle variables
            as linear combinations of the planets' orbital elements.
        new_qp_pairs : list, optional
            A list of the symbols to use for the newly-generated
            coordinate-momentum pairs. List entries should be given in the form
            of 2-tuples containing a coordinate symbol followed by its
            conjugate momentum variable. If no list is supplied, the default
            symbols :math:`(Q_i,P_i)` will be used.

        Returns
        -------
        .CanonicalTransformation
            The resulting transformation.
        """
        if not type(p_vars) == Poincare:
            raise TypeError("'p_vars' must be of type 'Poincare'")
        try:
            N_dof = p_vars.N_dof
            Tmtrx = np.array(Tmtrx).reshape(N_dof,N_dof)
        except:
            raise ValueError("'Tmtrx' could not be shaped into a {0}x{0} array".format(N_dof))
        Tmtrx = Matrix(Tmtrx)
        if not new_qp_pairs:
            new_qp_pairs = _get_default_qp_symbols(N_dof)
        Npl = p_vars.N - 1
        old_varslist = p_vars.qp_vars
        old_angvars = [old_varslist[3*i + 0] for i in range(Npl)] # lambdas
        old_angvars += [atan2(old_varslist[3*i + 1],old_varslist[3*(i+Npl) + 1]) for i in range(Npl)] # gammas
        old_angvars += [atan2(old_varslist[3*i + 2],old_varslist[3*(i+Npl) + 2]) for i in range(Npl)] # qs
        old_actvars = [old_varslist[3*(i+Npl) + 0] for i in range(Npl)] # Lambdas
        old_actvars += [(old_varslist[3*i + 1]**2 + old_varslist[3*(i+Npl) + 1]**2) / 2 for i in range(Npl)] # Gammas
        old_actvars += [(old_varslist[3*i + 2]**2 + old_varslist[3*(i+Npl) + 2]**2) / 2 for i in range(Npl)] # Qs

        n2o = dict(zip([Q for Q,P in new_qp_pairs],Tmtrx * Matrix(old_angvars)))
        n2o.update(dict(zip([P for Q,P in new_qp_pairs],Tmtrx.inv().transpose() * Matrix(old_actvars))))
        
        o2n=dict()
        old2new_angvars = Tmtrx.inv() * Matrix([Q for Q,P in new_qp_pairs])
        old2new_actvars = Tmtrx.transpose() * Matrix([P for Q,P in new_qp_pairs])
        Lambdas,lambdas = [[old_varslist[3*i + N] for i in range(Npl)] for N in (3*Npl,0)]
        kappas,etas,sigmas,rhos = [[old_varslist[3*i + N] for i in range(Npl)] for N in (1 + 3*Npl,1,2 + 3*Npl,2)]

        o2n.update(dict(zip(Lambdas,old2new_actvars[:Npl])))
        o2n.update(dict(zip(lambdas,old2new_angvars[:Npl])))

        kappa_exprns = [sqrt(2 * old2new_actvars[i]) * cos(old2new_angvars[i]) for i in range(Npl,2*Npl)]
        eta_exprns = [sqrt(2 * old2new_actvars[i]) * sin(old2new_angvars[i]) for i in range(Npl,2*Npl)]
        sigma_exprns = [sqrt(2 * old2new_actvars[i]) * cos(old2new_angvars[i]) for i in range(2*Npl,3*Npl)]
        rho_exprns = [sqrt(2 * old2new_actvars[i]) * sin(old2new_angvars[i]) for i in range(2*Npl,3*Npl)]
        o2n.update(dict(zip(kappas,kappa_exprns)))
        o2n.update(dict(zip(etas,eta_exprns)))
        o2n.update(dict(zip(sigmas,sigma_exprns)))
        o2n.update(dict(zip(rhos,rho_exprns)))
       
        new_qp_vars = [Q for Q,P in new_qp_pairs] + [P for Q,P in new_qp_pairs]
        return cls(p_vars.qp_vars,new_qp_vars,o2n,n2o,old_to_new_simplify=_termwise_trigsimp)

    @classmethod
    def Lambdas_to_delta_Lambdas(cls,pham):
        r"""
        Generate a canonical transformation applicable to a 
        PoincareHamiltonian object, `pham`, that replaces 
        the canonical momenta :math:`\Lambda_i` conjugate to 
        the mean longitudes :math:`\lambda_i` with new momenta
        :math:`\delta\Lambda_i = \Lambda_i - \Lambda_{i,0}`.
        """
        Lambda0s = pham.Lambda0s[1:]
        N = pham.N
        Lambdas = symbols("Lambda(1:{})".format(N))
        dLambdas = symbols(r"\delta\Lambda_{{(1:{})}}".format(N))
        o2n = {L:L0+dL for L,L0,dL in zip(Lambdas,Lambda0s,dLambdas)}
        n2o = {dL:(L-L0) for L,L0,dL in zip(Lambdas,Lambda0s,dLambdas)}
        L2dL= dict(zip(Lambdas,dLambdas))
        qpnew = [v.subs(L2dL) for v in pham.qp_vars]
        qpold = pham.qp_vars
        o2n_full={v:v.subs(o2n) for v in pham.qp_vars}
        n2o_full={v:v.subs(n2o) for v in qpnew}
        params = {L0:pham.H_params[L0] for L0 in Lambda0s}
        return cls(qpold,qpnew,o2n_full,n2o_full,params = params)
    
    @classmethod
    def actions_to_delta_actions(cls, old_qp_vars, actions, delta_actions, actions_ref, params={}):
        r"""
        Generate a canonical transformation applicable to a 
        PoincareHamiltonian object, `pham`, that replaces 
        the canonical momenta :math:`\Lambda_i` conjugate to 
        the mean longitudes :math:`\lambda_i` with new momenta
        :math:`\delta\Lambda_i = \Lambda_i - \Lambda_{i,0}`.
        """
        o2n = {a:a0+da for a,a0,da in zip(actions,actions_ref,delta_actions)}
        n2o = {da:(a-a0) for a,a0,da in zip(actions,actions_ref,delta_actions)}
        a2da= dict(zip(actions,delta_actions))
        qpnew = [v.subs(a2da) for v in old_qp_vars]
        qpold = old_qp_vars
        o2n_full={v:v.subs(o2n) for v in old_qp_vars}
        n2o_full={v:v.subs(n2o) for v in qpnew}
        return cls(qpold,qpnew,o2n_full,n2o_full,params=params)


    @classmethod
    def rescale_transformation(cls,qp_pairs, scale, cartesian_pairs = [], **kwargs):
        r"""
        Get a canonical transformation that simulatneously rescales the Hamiltonian
        and canonical momenta by a common factor. 

        Arguments
        ---------
        qp_pairs : list of 2-tuples
            Pairs of canonically conjugate variable symbols.
        scale : symbol or real
            Re-scaling factor. 
            The new momenta will be given by p' = scale * p
            and the new Hamiltonian will be H' = scale * H.
        cartesian pairs : list
            List of indices of Cartesian-style canonical pairs.
            These pairs will be recscaled such that 
            (y',x') = sqrt(scale) * (y,x)
        """
        o2n = dict()
        n2o = dict()
        rtscale = sqrt(scale)
        qpvars = [q for q,p in qp_pairs] + [p for q,p in qp_pairs]
        for i,qp in enumerate(qp_pairs):
            q,p = qp
            if i in cartesian_pairs:
                o2n[p] = p / rtscale 
                o2n[q] = q / rtscale 
                n2o[p] = p * rtscale 
                n2o[q] = q * rtscale 
            else:
                o2n[p] = p / scale 
                n2o[p] = p * scale 
        return cls(qpvars,qpvars,o2n,n2o,H_scale = scale,**kwargs)

    @classmethod
    def Poincare_rescale_transformation(cls,pham,scale,**kwargs):
        N = pham.N
        cart_pairs = [3*i + 1 for i in range(N-1)] + [3*i + 2 for i in range(N-1)]
        return cls.rescale_transformation(pham.qp_pairs,scale,cart_pairs,**kwargs)

    @classmethod
    def composite(cls, transformations, old_to_new_simplify = None, new_to_old_simplify = None):
        r"""
        Create a canonical transformation as the composition of canonical
        transformations.

        Arguments
        ---------
        transformations : list of :class:`~celmech.canonical_transformations.CanonicalTransformation`
            List of transformations to compose. Transformations are composed in
            the order

            .. math::
                T_N \circ T_{N-1}\circ ...\circ T_1 \circ T_0

            when passed the list :math:`[T_0,T_1,...,]`.

        Returns
        -------
        .CanonicalTransformation
        """
        old_qp_vars = transformations[0].old_qp_vars
        new_qp_vars = transformations[-1].new_qp_vars
        scale = Mul(*[t.H_scale for t in transformations])
        params = dict()
        n2o = dict()
        for var in new_qp_vars:
            var_exprn = var
            for t in reversed(transformations):
                var_exprn = t.new_to_old(var_exprn)
            n2o[var] = var_exprn
        o2n = dict()
        for var in old_qp_vars:
            var_exprn = var
            for t in transformations:
                var_exprn = t.old_to_new(var_exprn)
            o2n[var] = var_exprn
        for t in transformations:
            params.update(t.params)
    
        return cls(old_qp_vars, new_qp_vars, o2n, n2o, old_to_new_simplify, new_to_old_simplify,H_scale=scale,params=params)



import numpy as np
from ctypes import *
from .disturbing_function import  df_coefficient_C,evaluate_df_coefficient_dict,list_resonance_terms
#from . import clibcelmech
from celmech.disturbing_function import df_arguments_dictionary
#libname = "/Users/shadden/Projects/celmech/src/libcelmech.so"
#clibcelmech = CDLL(libname)
from . import clibcelmech
_rt2 = np.sqrt(2)
_rt2_inv = 1  / _rt2

class SeriesTerm(Structure):
    pass
SeriesTerm._fields_ = [
            ("k",c_int * 6),
            ("z",c_int * 4),
            ("coeff",c_double),
            ("next",POINTER(SeriesTerm))
            ]

_evaluate_series = clibcelmech.evaluate_series
_evaluate_series.argtypes = [
    np.ctypeslib.ndpointer(dtype = np.complex128,ndim=1),
    np.ctypeslib.ndpointer(dtype = np.complex128,ndim=1),
    POINTER(SeriesTerm),
    c_int,
    c_int,
    POINTER(c_double),
    POINTER(c_double)
]
_evaluate_series.restype = None

_evaluate_series_and_derivs = clibcelmech.evaluate_series_and_derivs
_evaluate_series_and_derivs.argtypes = [
    np.ctypeslib.ndpointer(dtype = np.complex128,ndim=1),
    np.ctypeslib.ndpointer(dtype = np.complex128,ndim=1),
    POINTER(SeriesTerm),
    c_int,
    c_int,
    POINTER(c_double),
    POINTER(c_double),
] + [(4 * c_double) for _ in range(4)]
_evaluate_series_and_derivs.restype = None

_evaluate_series_and_jacobian = clibcelmech.evaluate_series_and_jacobian
_evaluate_series_and_jacobian.argtypes = [
    np.ctypeslib.ndpointer(dtype = np.complex128,ndim=1),
    np.ctypeslib.ndpointer(dtype = np.complex128,ndim=1),
    POINTER(SeriesTerm),
    c_int,
    c_int,
    POINTER(c_double),
    POINTER(c_double),
] + [(4 * c_double) for _ in range(4)] + [(64 * c_double) for _ in range(2)]
_evaluate_series_and_jacobian.restype = None

def get_generic_df_coefficient_symbol(k1,k2,k3,k4,k5,k6,z1,z2,z3,z4):
    return symbols("C_{0}\,{1}\,{2}\,{3}\,{4}\,{5}^{6}\,{7}\,{8}\,{9}".format(
        k1,k2,k3,k4,k5,k6,z1,z2,z3,z4)
    )
from sympy import symbols,exp,I
def get_term_symbol(k1,k2,k3,k4,k5,k6,z1,z2,z3,z4):
    X,Xprime,Y,Yprime = symbols('X,Xprime,Y,Yprime')
    Xbar,Xprimebar,Ybar,Yprimebar = symbols('Xbar,Xbarprime,Ybar,Ybarprime')
    l,lprime = symbols('lambda,lambdaprime')
    coeff = get_generic_df_coefficient_symbol(k1,k2,k3,k4,k5,k6,z1,z2,z3,z4)

    term = coeff * exp(I * (k1 * lprime + k2 * l))
    term *= X**k3 if k3>0 else Xbar**(-k3)
    term *= Xprime**k4 if k4>0 else Xprimebar**(-k4)
    term *= Y**k5 if k5>0 else Ybar**(-k5)
    term *= Yprime**k6 if k6>0 else Yprimebar**(-k6)
    for z,prod in zip((z1,z2,z3,z4), (Y*Ybar,Yprime*Yprimebar,X*Xbar,Xprime*Xprimebar)):
        if z>0:
            term *= prod**z
    return term

class DFTermSeries(object):
    def __init__(self,resterm_dictionary,Lambda0In,Lambda0Out):
        r"""
        An object representing a Poisson series in variables
         exp[i\lambda_out],exp[i\lambda_in],Xin,Xout,Yin, and Yout.

        An instance can be used to calculate the value of the 
        Poisson series as well as its first and second derivatives
        with respect to Xin,Xout,Yin,and Yout.

        Arguments
        ---------
        resterm_dictionary : dict
          Supply the terms of the Poisson series and their coefficients.
          Dictionary entries are in the form {(kvec,zvec):Coeff}          
        Lambda0In : float
          The value of the inner planet's canonical momentum, Lambda,
          which is trated as consant in the Poisson series. 
          The value is neccessary for computing the value of derivatives
          of the canonical variables \eta,\kappa,\rho,\sigma.
        Lambda0Out : float
          The value of the outer planet's canonical momentum, Lambda,
          which is trated as consant in the Poisson series. 
          The value is neccessary for computing the value of derivatives
          of the canonical variables \eta,\kappa,\rho,\sigma.
        """
        self.slast_pointer = None
        self.s_dXbar_last_pointer = None
        kmax = 0
        Nmax = 0
        for kzpair,coeff_value in resterm_dictionary.items():
            ks,zs = kzpair
            kmax = max(kmax,abs(ks[0]),abs(ks[1]))
            Nmax = max(
                      Nmax,
                      abs(ks[2]) + zs[2],
                      abs(ks[3]) + zs[3],
                      abs(ks[4]) + zs[0],
                      abs(ks[5]) + zs[1]
                    )
            s = SeriesTerm()
            s.k = (6 * c_int)(*ks)
            s.z = (4 * c_int)(*zs)
            s.coeff = coeff_value
            s.next = self.slast_pointer
            self.slast_pointer = pointer(s)
        self.s0 = s
        self.Nmax = Nmax
        self.kmax = kmax
        rtLmbdaInv = 1 / np.sqrt([Lambda0In,Lambda0Out])
        mtrx = np.diag( np.concatenate((rtLmbdaInv, 0.5 * rtLmbdaInv )) )
        self.dXY_dQP  = np.block([
            [-1j * mtrx, +1j * mtrx],
            [mtrx, mtrx]
            ])
        Zeros = np.zeros((4,4))
        Id = np.eye(4)
        self.Omega = np.block([[Zeros,Id],[-Id,Zeros]])

    @classmethod
    def from_resonance_list(cls,resterm_list,G,mIn,mOut,MIn,MOut,Lambda0In,Lambda0Out):
        muIn = mIn * (MIn - mIn) / MIn
        muOut = mOut * (MOut - mOut) / MOut
        aIn0 = (Lambda0In / muIn)**2 / MIn / G
        aOut0 = (Lambda0Out / muOut)**2 / MOut / G
        alpha0 = aIn0 / aOut0
        aOut_inv = G*MOut*muOut*muOut / Lambda0Out / Lambda0Out  
        prefactor = -G * mIn * mOut * aOut_inv
        assert alpha0 < 1, "Particles are not in order by semi-major axis."
        resterm_dictionary  = {
                (ks,zs):prefactor * evaluate_df_coefficient_dict(df_coefficient_C(*ks,*zs),alpha0)
                for ks,zs in resterm_list
                }
        return cls(resterm_dictionary,Lambda0In,Lambda0Out)
    @classmethod
    def from_resonance_range(cls,j,k,Nmin,Nmax,G,mIn,mOut,MIn,MOut,Lambda0In,Lamda0Out):
        terms = list_resonance_terms(j,k,Nmin,Nmax)
        return cls.from_resonance_list(terms,G,mIn,mOut,MIn,MOut,Lambda0In,Lamda0Out)

    def _evaluate(self,lambda_arr, xy_arr):
        expIL = np.exp( 1j * lambda_arr)
        sum_re,sum_im  = c_double(),c_double()
        _evaluate_series(
                expIL,
                xy_arr,
                self.s0,
                self.kmax,
                self.Nmax,
                pointer(sum_re),
                pointer(sum_im)
            )
        return np.float64(sum_re)

    def _evaluate_with_derivs(self,lambda_arr,xy_arr):
        expIL = np.exp( 1j * lambda_arr)
        sum_re,sum_im  = c_double(),c_double()
        dS_dxy_re,dS_dxy_im,dS_dxybar_re,dS_dxybar_im = [(4 * c_double)() for _ in range(4)]
        _evaluate_series_and_derivs(
                expIL,
                xy_arr,
                self.s0,
                self.kmax,
                self.Nmax,
                pointer(sum_re),
                pointer(sum_im),
                dS_dxy_re,dS_dxy_im,dS_dxybar_re,dS_dxybar_im
            )
        dSdxy = np.concatenate((
            np.array(dS_dxy_re) + 1j * np.array(dS_dxy_im),
            np.array(dS_dxybar_re) + 1j * np.array(dS_dxybar_im)
            ))
        derivs = np.real(self.Omega @ self.dXY_dQP @ dSdxy)
        return np.float64(sum_re),derivs

    def _evaluate_with_jacobian(self,lambda_arr,xy_arr):
        expIL = np.exp( 1j * lambda_arr)
        sum_re,sum_im  = c_double(),c_double()
        dS_dxy_re,dS_dxy_im,dS_dxybar_re,dS_dxybar_im = [(4 * c_double)() for _ in range(4)]
        jac_re,jac_im = [(64 * c_double)() for _ in range(2)]
        _evaluate_series_and_jacobian(
                expIL,
                xy_arr,
                self.s0,
                self.kmax,
                self.Nmax,
                pointer(sum_re),
                pointer(sum_im),
                dS_dxy_re,dS_dxy_im,dS_dxybar_re,dS_dxybar_im,
                jac_re, jac_im
            )

        dSdxy = np.concatenate((
            np.array(dS_dxy_re) + 1j * np.array(dS_dxy_im),
            np.array(dS_dxybar_re) + 1j * np.array(dS_dxybar_im)
            ))
        derivs = np.real(self.Omega @ self.dXY_dQP @ dSdxy)
        jac_xy = (np.array(jac_re) + 1j * np.array(jac_im)).reshape(8,8)
        jac_qp = self.Omega @ np.real(self.dXY_dQP @ jac_xy @ self.dXY_dQP.T)
        return np.float64(sum_re), derivs, jac_qp, jac_xy

    def PoincareParticlesEvaluate(self,pvars,indexIn,indexOut):
        ps = (pvars.particles[indexIn],pvars.particles[indexOut])
        xy = [(p.kappa - 1j * p.eta) / np.sqrt(p.Lambda) for p in ps]
        xy += [0.5 * (p.sigma - 1j * p.rho) / np.sqrt(p.Lambda) for p in ps]
        xy = np.array(xy)
        lambdas = np.array([p.l for p in ps])
        H,vardot,jac_qp,jac_xy = self._evaluate_with_jacobian(lambdas,xy)
        return {'Hamiltonain':H,'derivatives':vardot,'Jacobian':jac_qp}



import theano.tensor as T
from celmech.theano_ops.kepler import KeplerOp
def planar_els2xv(a,lmbda,h,k,GMstar):
    ko = KeplerOp()
    e_sq = h*h+k*k
    e = T.sqrt(e_sq)
    # ko(MeanAnomaly,eccentricity) = sin_TrueAnomly, cos_TrueAnomaly
    sin_f,cos_f =  ko( lmbda - T.arctan2(h,k), e + T.zeros_like(lmbda))
    cos_theta = cos_f * (k/e) - sin_f * (h/e)
    sin_theta = sin_f * (k/e) + cos_f * (h/e)
    n = T.sqrt(GMstar) * a**(-3/2)
    ecosf = k * cos_theta + h * sin_theta
    r = a * (1-e_sq) /(1 + ecosf)
    x = r * cos_theta
    y = r * sin_theta
    vel = n * a / T.sqrt(1-e_sq)
    u = -1 * vel * (h + sin_theta)
    v = vel * (k + cos_theta)
    return x,y,u,v

def R_inc_mtrx_transform(x,y,u,v,p,q):
    cosIby2 = T.sqrt(1 - p * p - q * q)
    
    x1 = (1-2*p*p) * x + 2 * p * q * y
    y1 = 2 * p * q * x + (1-2*q*q) * y
    z1 = -2 * p * cosIby2 * x + 2 * q * cosIby2 * y 

    u1 = (1-2*p*p) * u + 2 * p * q * v
    v1 = 2 * p * q * u + (1-2*q*q) * v
    w1 = -2 * p * cosIby2 * u + 2 * q * cosIby2 * v 

    return x1,y1,z1,u1,v1,w1

def calc_Hint_components_planar(a1,a2,l1,l2,h1,k1,h2,k2,GMstar1,GMstar2):
    r"""
    Compute the value of the disturbing function components
    .. math::
        H_{dir} = -\frac{1}{|r-r'|}
        H_{ind} = v.v'  
    from a set of input orbital elements for coplanar planets.

    Arguments
    ---------
    a1 : float
        inner planet semi-major axis
    a2 : float
        outer planet semi-major axis
    l1 : float
        inner planet mean longitude
    l2 : float
        outer planet mean longitude
    h1 :
        e1 * sin(pomega1)
    k1 : 
        e1 * cos(pomega1)
    h2 :
        e2 * sin(pomega2)
    k2 : 
        e2 * cos(pomega2)

    Returns
    -------
    float :
        Disturbing function value
    """
    x1,y1,u1,v1 = planar_els2xv(a1,l1,h1,k1,GMstar1)
    x2,y2,u2,v2 = planar_els2xv(a2,l2,h2,k2,GMstar2)

    # direct term
    dx = (x2 - x1)
    dy = (y2 - y1)
    dr2 = dx*dx + dy*dy
    direct = -1 / T.sqrt(dr2)

    # indirect term
    indirect = u1*u2 + v1*v2
    return direct,indirect

def calc_Hint_components_spatial(a1,a2,l1,l2,h1,k1,h2,k2,p1,q1,p2,q2,GMstar1,GMstar2):
    r"""
    Compute the value of the disturbing function components
    .. math::
        H_{dir} = -\frac{1}{|r-r'|}
        H_{ind} = v.v'  
    from a set of input orbital elements for coplanar planets.

    Arguments
    ---------
    a1 : float
        inner planet semi-major axis
    a2 : float
        outer planet semi-major axis
    l1 : float
        inner planet mean longitude
    l2 : float
        outer planet mean longitude
    h1 :
        e1 * sin(pomega1)
    k1 : 
        e1 * cos(pomega1)
    h2 :
        e2 * sin(pomega2)
    k2 : 
        e2 * cos(pomega2)
    p1 : float
        sin(I1/2) sin(Omega1)
    q1 : float
        sin(I1/2) cos(Omega1)
    p2 : float
        sin(I2/2) sin(Omega2)
    q2 : float
        sin(I2/2) cos(Omega2)

    Returns
    -------
    float :
        Disturbing function value
    """
    _x1,_y1,_u1,_v1 = planar_els2xv(a1,l1,h1,k1,GMstar1)
    _x2,_y2,_u2,_v2 = planar_els2xv(a2,l2,h2,k2,GMstar2)

    x1,y1,z1,u1,v1,w1 = R_inc_mtrx_transform(_x1,_y1,_u1,_v1,p1,q1)
    x2,y2,z2,u2,v2,w2 = R_inc_mtrx_transform(_x2,_y2,_u2,_v2,p2,q2)
    # direct term
    dx = (x2 - x1)
    dy = (y2 - y1)
    dz = (z2 - z1)
    dr2 = dx*dx + dy*dy + dz*dz
    direct = -1 / T.sqrt(dr2)

    # indirect term
    indirect = u1*u2 + v1*v2 + w1*w2
    return direct,indirect



from sympy import S, diff, lambdify, symbols, Matrix, Expr
try:
    from collections.abc import MutableMapping
except:
    from collections import MutableMapping
import pprint
from numpy import array
from collections import OrderedDict, UserDict
import numpy as np
from scipy.integrate import ode
import scipy.special
from .miscellaneous import poisson_bracket

def _my_elliptic_e(*args):
    if len(args) == 1:
        return scipy.special.ellipe(*args)
    else:
        return scipy.special.ellipeinc(*args)

_lambdify_kwargs = {'modules':['numpy', {
    'elliptic_k': scipy.special.ellipk,
    'elliptic_f': scipy.special.ellipkinc,
    'elliptic_e': _my_elliptic_e
    }]} 

# Should reduce hamiltonian only check for cyclic q, and then set both q and p as parameters? 
# There are cases where p shows up in hamiltonian but not q

# Do we call update in fullqp setter? Do we add a needs_update that only gets called on integrate(or others)?

# units in canonical transformation, Ham doesn't know about that

# All params and units are owned by Hamiltonians. PhaseSpaceStates are just numbers with no
# knowledge of units. 

# tracked and untracked variables handled by Hamiltonian
# we do want to be able to set things in phasespacestate from hamiltonian and update hamiltonian
# state is a property in hamiltonian that sets update if we call setter, and can handle the logic on 
# which phasepsacestate variables to return based on which dof are being tracked or not

# full values gives additional variables including conserved quantitties

class PhaseSpaceState(object):
    """
    A general class for describing the phase-space state
    of a Hamiltonian dynamical system.
    
    Attributes
    ----------
    qp : OrderedDict
        An ordered dictionary containing the canonical
        coordiantes and momenta as keys and their numerical
        values as dictionary values.
    qp_vars : list
        List of variable symbols used for the canonical
        coordinates and momenta.
    qp_pairs : list
        A list of the 2-tuples :math:`(q_i,p_i)`.
    N_dof : int
        The number of degrees of freedom.
    values : list
        List of the numerical values of `qp_vars`.
    """
    def __init__(self, qp_vars, values, t=0):
        """
        Arguments
        ---------
        qp_vars : list of symbols
            List of symbols to use for the canonical coordiantes
            and momenta. The list should be of length 2*N for an
            integer N with entries 0,...,N-1 representing the
            canonical coordiante variables and entries N,...,2*N-1
            representing the corresponding conjugate momenta.
        values : array-like
            The numerical values assigned to the canonical coordiantes
            and momenta.
        t : float, optional
            The current time of system represented by the phase-space
            state. Default value is 0.
        """
        self.t = t
        msg="'qp_vars' and 'values' must have the same dimension."
        assert len(qp_vars)==len(values), msg
        self.qp = qpDict(qp_vars, values)

    @property
    def qp_vars(self):
        return list(self.qp.keys()) 

    @property
    def qp_pairs(self):
        return [(self.qp_vars[i], self.qp_vars[i+self.N_dof]) for i in range(self.N_dof)]

    @property 
    def N_dim(self):
        return len(self.qp_vars)

    @property 
    def N_dof(self):
        return int(self.N_dim/2)

    @property
    def values(self):
        return list(self.qp.values()) 
    @values.setter
    def values(self,values):
        for key, value in zip(self.qp_vars, values):
            self.qp[key] = value

    def __str__(self):
        s = "t={0}".format(self.t)
        for var, val in self.qp.items():
            s += ", {0}={1}".format(var, val)
        return s
    def __repr__(self):
        return "PhaseSpaceState(qp_vars={0}, values={1}, t={2})".format(self.qp_vars, self.values, self.t)

class Hamiltonian(object):
    """
    A general class for describing and evolving Hamiltonian systems.

    Attributes
    ----------

    state : object
        An object that stores the dynamical state of the system.

    H : sympy expression
        Symbolic expression for system's Hamiltonian.

    pqpars : list
        List of canonical variable pairs. 
    """
    def __init__(self, H, H_params, state, full_qp_vars=None):
        """
        Arguments
        ---------
        H : sympy expression
            Hamiltonian made up only of sympy symbols in state.qp_pairs and keys in H_params
        H_params : dict
            dictionary from sympy symbols for the constant parameters in H to their value
        state : PhaseSpaceState 
            Object for holding the dynamical state.
        
        In addition to the above, one needs to write 2 methods to map between the two objects:
        def state_to_list(self, state): 
            returns a list of values from state in the same order as pqpairs e.g. [P1,Q1,P2,Q2]
        def update_state_from_list(self, state, y, t):
            updates state object from a list of values y for the variables in the same order as pqpairs
            and integrator time 't'
        """
        self._H_params = ParamDict(self, H_params)
        self._H = H
        self.state = state
        self._full_qp_vars = full_qp_vars
        self._needs_update = True
        
    @property
    def t(self):
        return self.state.t

   
    # Property so that user can't inadvertently replace it with a regular dictionary
    @property
    def H_params(self):
        return self._H_params
    
    @property 
    def N_dim(self):
        return self.state.N_dim
    
    @property 
    def full_N_dim(self):
        return len(self.full_qp_vars)

    @property 
    def N_dof(self):
        return self.state.N_dof
    
    @property 
    def full_N_dof(self):
        return int(self.full_N_dim/2)

    @property
    def qp(self):
        return self.state.qp
    
    @property
    def full_qp(self):
        full_qp = Fullqp(self)
        return full_qp
    
    @property
    def qp_pairs(self):
        return self.state.qp_pairs
    
    @property
    def full_qp_pairs(self):
        if self._full_qp_vars:
            return [(self._full_qp_vars[i], self.full_qp_vars[i+self.full_N_dof]) for i in range(self.full_N_dof)]
        else:
            return self.qp_pairs

    @property
    def qp_vars(self):
        return self.state.qp_vars
    
    @property
    def full_qp_vars(self):
        if self._full_qp_vars:
            return self._full_qp_vars
        else:
            return self.qp_vars
    
    @property
    def values(self):
        return self.state.values
    @values.setter
    def values(self,vals):
        self.state.values = values

    @property
    def full_values(self):
        return list(self.full_qp.values())
    @full_values.setter
    def full_values(self,vals):
        assert len(vals) == self.full_N_dim
        qpvars = self.qp_vars
        for var,val in zip(self.full_qp_vars,vals):
            if var in qpvars:
                self.state.qp[var] = val
            else:
                self.H_params[var] = val
    ##############
    # symbolic Hamiltonian, flow, and Jacobian
    ##############
    @property
    def H(self):
        return self._H
    @H.setter
    def H(self, H):
        self._H = H
        self._needs_update = True
    @property
    def flow(self):
        r"""Symbolic representation of the flow,
        :math:`(\frac{\partial}{\partial p}H,-\frac{\partial}{\partial q}H)`"""
        if self._needs_update:
            self._update()
        return self._flow
    @property
    def jacobian(self):
        r"""Symbolic representation of the Jacobian of the flow."""
        if self._needs_update:
            self._update()
        return self._jacobian
    ##############
    # numerical Hamiltonian, flow, and Jacobian
    ##############
    @property
    def N_H(self):
        if self._needs_update:
            self._update()
        return self._N_H
    @property
    def N_flow(self):
        if self._needs_update:
            self._update()
        return self._N_flow
    @property
    def N_jacobian(self):
        if self._needs_update:
            self._update()
        return self._N_jacobian
    ##############
    # functions for Hamiltonian, flow, and Jacobian
    ##############
    @property
    def H_func(self):
        r"""Hamiltonian function, taking canonical variables as arguments."""
        if self._needs_update:
            self._update()
        return self._H_func
    @property
    def flow_func(self):
        r"""Fucntion of canonical variables that returns the Hamiltonian flow
        vector."""
        if self._needs_update:
            self._update()
        return self._flow_func
    @property
    def jacobian_func(self):
        r"""Function of canonical variables that returns the Jacobian of the
        Hamiltonian flow with respect to the canonical variables."""
        if self._needs_update:
            self._update()
        return self._jacobian_func
    ##################
    # evaluate Hamiltonian, flow, and jacobian at current state
    #################
    def calculate_H(self):
        """
        Calculate the Hamiltonian of the system
        in its current state.

        Arguments
        ---------
        None

        Returns
        -------
        energy : float
            The numerical value of the Hamiltonian evaluated at
            the current phase space state of the system.
        """
        energy = self.H_func(*self.values)
        return energy
    # Alias for calculate H
    calculate_energy = calculate_H
    def calculate_flow(self):
        """
        Calculate the flow vector for the system
        in its current state.

        Arguments
        ---------
        None

        Returns
        -------
        flow : ndarray, shape (N,)
            The numerical value of the flow vector evaluated at
            the current phase space state of the system.
            N is twice the number of degrees of freedom of the system.
        """
        flow = self.flow_func(*self.values).reshape(-1)
        return flow
    def calculate_jacobian(self):
        """
        Calculate the jacobian matrix of the equations of motion
        for the system in its current state.

        Arguments
        ---------
        None

        Returns
        -------
        jacobian : ndarray, shape (N,N)
            The numerical value of the Jacobian matrix evaluated at
            the current phase space state of the system.
            N is twice the number of degrees of freedom of the system.
        """
        jac = self.jacobian_func(*self.values)
        return jac

    @property
    def integrator(self):
        if self._needs_update:
            self._update()
        return self._integrator
     
    def Lie_deriv(self,exprn):
        r"""
        Return the Lie derivative of an expression with respect to the Hamiltonian.
        In other word, compute 

        .. math::
            \mathcal{L}_{H}f

        where :math:`f` is the argument passed and :math:`\mathcal{L}_H \equiv [\cdot,H]` 
        is the Lie derivative operator.

        Arguments
        ---------
        exprn : sympy expression
            The expression to take the Lie derivative of.

        Returns
        -------
        sympy expression
            sympy expression for the resulting derivative.
        """
        return poisson_bracket(exprn,self.H,self.qp_vars,[])

    def N_Lie_deriv(self,exprn):
        r"""
        Return the Lie derivative of an expression with respect to the Hamiltonian
        with numerical values substituted for parameters. Equivalent to 
        :meth:`~poincare.Hamiltonian.N_Lie_deriv` but using the NH attribute 
        rather than the H attribute to compute brackets.  

        Arguments
        ---------
        exprn : sympy expression
            The expression to take the Lie derivative of.

        Returns
        -------
        sympy expression
            sympy expression for the resulting derivative.
        """
        return poisson_bracket(exprn,self.N_H,self.qp_vars,[])
    def set_integrator(self,name,**integrator_params):
        """
        Set the integrator and corresponding integration parameters. This
        method provides a wrapper to the 
        :meth:`scipy.integrate.ode.set_integrator` method.

        See `here <https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.ode.set_integrator.html#scipy.integrate.ode.set_integrator>`_ for additional details.

        Arguments
        ---------
        name : str
            Name of the integrator.
        integrator_params :
            Additional parameters for the integrator.
        """
        if self._needs_update:
            self._update()
        # Override scipy's default relative tolerance 
        integrator_params.setdefault('rtol',1e-14)
        self._integrator.set_integrator(name,**integrator_params)

    def integrate(self, time):
        """
        Evolve Hamiltonian system from current state
        to input time by integrating the equations of 
        motion.

        Arguments
        ---------
        time : float
            Time to advance Hamiltonian system to.
        """
        # Sync the integrator time and values with what's in self.state in case user has changed it
        if self._needs_update:
            self._update()
        self.integrator.set_initial_value(y=self.state.values, t=self.state.t)
        try:
            self.integrator.integrate(time)
        except:
            raise AttributeError("Need to initialize Hamiltonian")
        # Sync self.state with outcome of integration
        self.state.t = self.integrator.t
        self.state.values = self.integrator.y

    def _update(self):
        self._needs_update=False # reset flag up top to avoid infinite recursion
        self._N_H = self._H # reset to Hamiltonian with all parameters unset
        # 
        # Update raw numerical constants first then update functions
        # Less hacky way to do this?
        function_keyval_pairs = []
        for key, val in self._H_params.items(): 
            if isinstance(val,float):
                self._N_H = self._N_H.subs(key, val)
            else:
                function_keyval_pairs.append((key,val)) 
        for keyval in function_keyval_pairs:
            self._N_H = self._N_H.subs(keyval[0],keyval[1])

        qp_vars = self.qp_vars
        flow = []
        Nflow = []
        for v in self.qp_vars:
            deriv = self.Lie_deriv(v)
            Nderiv = self.N_Lie_deriv(v)
            flow.append(deriv)
            Nflow.append(Nderiv)

        N_dim = 2*self.N_dof
        self._flow = Matrix(flow)
        self._N_flow = Matrix(Nflow)
        self._jacobian = Matrix(N_dim,N_dim, lambda i,j: diff(flow[i],qp_vars[j]))
        self._N_jacobian = Matrix(N_dim,N_dim, lambda i,j: diff(Nflow[i],qp_vars[j]))

        self._H_func = lambdify(qp_vars,self._N_H,**_lambdify_kwargs)
        self._flow_func = lambdify(qp_vars,self._N_flow,**_lambdify_kwargs)
        self._jacobian_func = lambdify(qp_vars,self._N_jacobian,**_lambdify_kwargs)

        self._integrator = ode(
                lambda t,y: self._flow_func(*y),
                jac = lambda t,y: self._jacobian_func(*y))
        self._integrator.set_integrator('vode',method='adams',rtol=1e-14)

def reduce_hamiltonian(ham,retain_explicit=[]):
    r"""
    Given a :class:`~celmech.hamiltonian.Hamiltonian` object, generate a new
    :class:`~celmech.hamiltonian.Hamiltonian` object with fewer degrees of
    freedom by determining which (if any) canonical variables do not appear
    explicitly in the Hamiltonian. 

    Arguments
    ---------
    ham : Hamiltonian
        The original Hamiltonian to reduce.
    retain_explicit: list
        List of variables for which to retain explicit dependence on in the
        transformed Hamiltonian. List entries should be either indices or
        variable symbols.

    Returns
    -------
    rham : Hamiltonian
        The reduced Hamiltonian.
    """
    state = ham.state
    new_params = ham.H_params.copy()
    untracked_q, untracked_p = [], []
    new_q, new_p = [], []
    new_qvals, new_pvals = [], []
    qp_pairs = state.qp_pairs
    for i,qp_pair in enumerate(qp_pairs):
        q,p = qp_pair
        retain_explicitQ = (i in retain_explicit) or (q in retain_explicit) or\
            (p in retain_explicit)
        pval,qval = state.qp[p],state.qp[q]
        # if q is cyclic, p is conserved or vice versa and we ignore this dof 
        dof_doesnt_appearQ = q not in ham.H.free_symbols or p not in ham.H.free_symbols
        if dof_doesnt_appearQ and not retain_explicitQ:
            untracked_q.append(q)
            untracked_p.append(p)
            new_params[q] = qval
            new_params[p] = pval
        else:
            new_q.append(q)
            new_p.append(p)
            new_qvals.append(qval)
            new_pvals.append(pval)
    new_vals = np.array(new_qvals + new_pvals)
    new_qp_vars = new_q + new_p
    untracked_qp_vars = untracked_q + untracked_p
    new_state = PhaseSpaceState(new_qp_vars, new_vals,state.t)
    new_ham = Hamiltonian(ham.H,new_params,new_state, full_qp_vars = ham.full_qp_vars)
    return new_ham

class qpDict(MutableMapping):
    def __init__(self, qp_vars, values):
        self._qp = OrderedDict(zip(qp_vars, values))

    def __getitem__(self, key):
        if isinstance(key, int):
            if key >= len(self._qp) or key < -len(self._qp):
                raise AttributeError("Accessing qp dictionary with an index ({0}) that is out of bounds".format(key))
            symbolkey = list(self._qp.keys())[key]
            return self._qp[symbolkey]
        else:
            return self._qp[key]

    def __setitem__(self, key, value):
        if isinstance(key, int):
            if key >= len(self._qp) or key < -len(self._qp):
                raise AttributeError("Setting qp dictionary with an index ({0}) that is out of bounds".format(key))
            symbolkey = list(self._qp.keys())[key]
            self._qp[symbolkey] = value
        else:
            self._qp[key] = value

    def __delitem__(self, key):
        if isinstance(key, int):
            if key >= len(self._qp) or key < -len(self._qp):
                raise AttributeError("Deleting item in qp dictionary with an index ({0}) that is out of bounds".format(key))
            symbolkey = list(self._qp.keys())[key]
            del self._qp[symbolkey]
        else:
            del self._qp[key]

    def __iter__(self):
        return iter(self._qp)

    def __len__(self):
        return len(self._qp)

    def __repr__(self):
        return repr(self._qp)

class Fullqp(MutableMapping):
    def __init__(self, hamiltonian):
        self.hamiltonian = hamiltonian

    def __getitem__(self, key):
        try: # first try to find in the dynamical variables qp, if not, look for conserved quantities in H_params (in case of reduced hamiltonian)
            return self.hamiltonian.qp[key]
        except:
            try:
                return self.hamiltonian.H_params[key]
            except:
                raise AttributeError('Variable {0} not found'.format(key))

    def __setitem__(self, key, value):
        try: # first try to find in the dynamical variables qp, if not, look for conserved quantities in H_params (in case of reduced hamiltonian)
            self.hamiltonian.qp[key] = value
        except:
            try:
                self.hamiltonian.H_params[key] = value
            except:
                raise AttributeError('Variable {0} not found'.format(key))

    def __delitem__(self, key):
        raise AttributeError("deleting variables not implemented.")

    def __iter__(self):
        for key in self.hamiltonian.full_qp_vars:
            yield key

    def __len__(self):
        return len(self.hamiltonian.full_qp_vars)

class ParamDict(MutableMapping):
    def __init__(self, hamiltonian, params):
        self.hamiltonian = hamiltonian
        self._params = params.copy()

    def copy(self):
        return ParamDict(self.hamiltonian, self._params)

    def __getitem__(self, key):
        return self._params[key]
    
    def __setitem__(self, key, value):
        self.hamiltonian._needs_update = True
        self._params[key] = value
    
    def __delitem__(self, key):
        del self._params[key]

    def __iter__(self):
        return iter(self._params)

    def __len__(self):
        return len(self._params)
    
    def __repr__(self):
        return repr(self._params)



import numpy as np
from .miscellaneous import sk,Dsk

class StandardMap():
    r"""
    A class representing the `Chirikov standard map`_.
    The map depends on a single parameter, :math:`K` 
    and is defind by

    .. math::
        \begin{align}
            p' &=& p + K  \sin\theta\\
            \theta' &=& \theta + p'
        \end{align}

    By default, the map is defined on the cylinder with
    the :math:`\theta` coordinate taken mod :math:`2\pi`.
    The parameter `mod_p=True` can be set to take the 
    :math:`p` coordinate modulo :math:`2\pi` as well.

    .. _Chirikov standard map: https://en.wikipedia.org/wiki/Standard_map

    Parameters
    ----------
    K : float
        Map non-linearity parameter.
    mod_theta : bool, optional
        If True, the :math:`\theta` coordinate
        is taken modulo :math:`2\pi`.
        Default is `True`
    mod_p : bool, optional
        If True, the :math:`p` coordinate
        is taken modulo :math:`2\pi`.
        Default is `False`.
    """
    def __init__(self,K,mod_theta = True,mod_p = False):
        self.K = K 
        self._mod_theta = mod_theta
        self._mod_p = mod_p
        self._set_modfn()

    @property
    def mod_theta(self):
        r"""
        Is the coordinate :math:`\theta` calculated
        modulo :math:`2\pi`?
        """
        return self._mod_theta
    @property
    def mod_p(self):
        r"""
        Is the coordinate :math:`p` calculated
        modulo :math:`2\pi`?
        """
        return self._mod_p
    @mod_theta.setter
    def mod_theta(self,value):
        self._mod_theta = value
        self._set_modfn()
    @property
    def mod_p(self):
        return self._mod_p
    @mod_p.setter
    def mod_p(self,value):
        self._mod_p = value
        self._set_modfn()


    def _set_modfn(self):
        iden = lambda x: x
        mod2pi = lambda x: np.mod(x,2*np.pi)
        
        pfn = mod2pi if self._mod_p else iden
        thetafn = mod2pi if self._mod_theta else iden
        self._modfn = lambda x: np.array([thetafn(x[0]),pfn(x[1])])
    
    def __call__(self,x):
        theta,p = x
        p1 = p + self.K * np.sin(theta)
        theta1 = theta + p1
        x1 = np.array([theta1,p1])
        x1 = self._modfn(x1)
        return x1

    def inv(self,x):
        r"""
        The inverse mapping

        .. math::
            \begin{align}
            \theta &=& p' - \theta' \\
            p &=& p' - K \sin\theta
            \end{align}


        Arguments
        ---------
        x : array-like
            The point :math:`(\theta',p')`

        Returns
        -------
        array-like
            The point :math:`(\theta,p)`
        """
        theta1,p1 = x
        theta = theta1 - p1
        p = p1 - self.K * np.sin(theta)

    def partial_derivs(self,x,Nmax):
        r"""
        Get the partial derivatives of
        the map evaluated at the point
        `x0` up to order `Nmax`.

        Arguments
        ---------
        x : array-like
            The point at which derivatives
            are to be evaluated
        Nmax : int
            Maximum order of the partial
            derivatives to return

        Returns
        -------
        T : array, shape (2,Nmax+1,Nmax+1)
            The partial derivatives of the map.
            Writing the value of the map at a point
            :math:`(x_1,x_2)` as 
            :math:`T(x_1,x_2) = (T_1(x_1,x_2),T_2(x_1,x_2))`,
            the entry T[i,n,m] stores
            .. math::
                \frac{\partial^{(n+m)}}{\partial x_1^n \partial x_2^m} T_i

            Note that ``T[:,0,0]`` give the value of the map.
        """
        theta,p = x
        c,s = np.cos(theta),np.sin(theta)
        K = self.K
        Ksin_derivs = K*np.array([s,c,-s,-c])
        T = np.zeros((2,Nmax+1,Nmax+1))
        T[:,0,0] = self.__call__(x)
        for n in range(1,Nmax+1):
            T[:,n,0] = Ksin_derivs[n%4]
        T[0,1,0]+=1
        T[0,0,1]+=1
        T[1,0,1]+=1
        return T

    def inv_partial_derivs(self,x,Nmax):
        r"""
        Get the partial derivatives of the inverse map evaluated at the point
        `x0` up to order `Nmax`.

        Arguments
        ---------
        x : array-like
            The point at which derivatives are to be evaluated
        Nmax : int
            Maximum order of the partial derivatives

        Returns
        -------
        T : array, shape (2,Nmax+1,Nmax+1)
            The partial derivatives of the map.  Writing the value of the map
            at a point :math:`(x_1,x_2)` as :math:`T(x_1,x_2) =
            (T_1(x_1,x_2),T_2(x_1,x_2))`, the entry T[i,n,m] stores

            .. math::
                \frac{\partial^{(n+m)}}{\partial x_1^n \partial x_2^m} T_i

            Note that T[:,0,0] give the value of the map.
        """
        theta1,p1 = x
        theta,p = self.inv(x)
        c,s = np.cos(theta),np.sin(theta)
        K = self.K
        Ksin_derivs = K*np.array([s,c,-s,-c])
        T = np.zeros((2,Nmax+1,Nmax+1))
        T[:,0,0] = theta,p
        T[0,1,0] = 1
        T[0,0,1] = -1
        for n in range(1,Nmax+1):
            for l in range(0,n+1):
                T[1,l,n-l] = -(-1)**(n-l) * Ksin_derivs[n%4]
        T[1,0,1]+=1
        return T

    def jac(self,x):
        r"""
        Evaluate the Jacobian map at :math:`x=(\theta,p)`, given by

        .. math::
            DT(x) = \begin{pmatrix}
                        1 + K\cos\theta & 1 \\
                        K\cos\theta     & 1
                    \end{pmatrix}
        
        Aruments
        --------
        x : array
            Point(s) at which to evaluate the Jacobian.

        Returns
        -------
        DT : array
            Value of the Jacobain at point x.
        """
        theta,p = x
        K = self.K
        Kcos = K*np.cos(theta)
        jac = np.array([
            [1 + Kcos, 1],
            [Kcos,1]
        ])
        return jac

class EncounterMap():
    r"""
    A class representing the encounter map.
    The map depends on three parameters,
    :math:`\epsilon,y`, and :math:`J`.
    The map is defined by the equations

    .. math::
        \begin{align}
        x' &= x + \epsilon f(\theta;y)
        \\
        \theta' &= \theta + 2\pi(J-x')
        \end{align}

    By default, the map is defined on the cylinder with
    the :math:`\theta` coordinate taken mod :math:`2\pi`.
    The parameter `mod_p=True` can be set to take the 
    :math:`p` coordinate modulo :math:`2\pi` as well.

    .. _Chirikov standard map: https://en.wikipedia.org/wiki/Standard_map

    Parameters
    ----------
    K : float
        Map non-linearity parameter.
    mod_theta : bool, optional
        If True, the :math:`\theta` coordinate
        is taken modulo :math:`2\pi`.
        Default is `True`
    mod_p : bool, optional
        If True, the :math:`p` coordinate
        is taken modulo :math:`2\pi`.
        Default is `False`.
    """
    def __init__(self,m,J,y0, Nmax=7, mod = True):
        self.m = m
        self.J = J
        self._Nmax = Nmax
        self._y0 = y0
        self._update_amplitudes()
        self.mod = mod

    @property
    def mod(self):
        return self._mod

    @mod.setter
    def mod(self,val):
        self._mod = val
        if val:
            self._modfn = lambda x: np.mod(x,2*np.pi)
        else:
            self._modfn = lambda x: x

    @property
    def J(self):
        return self._J

    @property
    def eps(self):
        da = self.da0
        m = self.m
        da4 = da*da*da*da
        return 8 * m /da4 / 9

    @J.setter
    def J(self,value):
        self._J = value
        self.da0 = 2/3/self._J

    @property
    def Nmax(self):
        return self._Nmax
    @property
    def y0(self):
        return self._y0

    @y0.setter
    def y0(self,value):
        self._y0 = value
        self._update_amplitudes()

    @Nmax.setter
    def Nmax(self,value):
        self._Nmax = value
        self._update_amplitudes()

    def _update_amplitudes(self):
        y0 = self._y0
        Nmax = self._Nmax
        self.amps = np.array([-4 * np.pi * k * sk(k,y0) / 3 for k in np.arange(1,Nmax+1)])

    def f(self,theta):
        sin_ktheta = np.array([np.sin(k*theta) for k in range(1,self.Nmax+1)])
        return self.amps @ sin_ktheta

    def dfdtheta_n(self,theta,n):
        trig = np.array([k**(n) * np.sin(k*theta + 0.5 * n * np.pi) for k in range(1,self.Nmax+1)])
        return self.amps @ trig

    def __call__(self,X):
        theta,x = X
        eps = self.eps
        x1 = x + eps * self.f(theta)
        theta1 = theta - 2 * np.pi * x1
        theta1 = self._modfn(theta1)
        return np.array([theta1,x1])

    def jac(self,X):
        theta,x = X
        dx1_dx = 1
        dx1_dtheta = self.eps * self.dfdtheta_n(theta,1)
        dtheta1_dx = -2*np.pi * dx1_dx
        dtheta1_dtheta = 1 - 2 * np.pi * dx1_dtheta
        return np.array([[dtheta1_dtheta,dtheta1_dx],[dx1_dtheta,dx1_dx]])

    def inv(self,X):
        theta1,x1 = X
        eps = self.eps
        theta = theta1 + 2 * np.pi * x1 
        x = x1 - eps * self.f(theta)
        return (theta,x)

    def partial_derivs(self,x0,Nmax):
        """
        Get the partial derivatives of the map up 
        to order `Nmax` evaluated at point `x0`.
        """
        theta,x = x0
        T = np.zeros((2,Nmax+1,Nmax+1))
        eps = self.eps
        T[:,0,0] = self.__call__(x0)
        T[0][0,1] = -2 * np.pi
        T[1][0,1] = 1
        n=1
        eps_fn = eps * self.dfdtheta_n(theta,n)
        T[0][1,0] = 1 - 2 * np.pi * eps_fn
        T[1][1,0] = eps_fn
        for n in range(2,Nmax+1):
            eps_fn = eps * self.dfdtheta_n(theta,n)
            T[0][n,0] = -2 * np.pi * eps_fn
            T[1][n,0] = eps_fn
        return T

    def inv_partial_derivs(self,x0,Nmax):
        """
        Get the partial derivatives of the map up 
        to order `Nmax` evaluated at point `x0`.
        """
        theta1,x1 = x0
        T = np.zeros((2,Nmax+1,Nmax+1))
        eps = self.eps
        T[:,0,0] = self.inv(x0)
        T[0][1,0] = 1
        T[0][0,1] = 2 * np.pi
        theta,x = T[:,0,0]
        for n in range(1,Nmax+1):
            eps_fn = eps * self.dfdtheta_n(theta,n)
            for l in range(n+1):
                T[1][l,n-l] = -1 * (2*np.pi)**(n-l) * eps_fn
        T[1][0,1] += 1
        return T


from sympy import bell
from scipy.special import binom
def _evaluate_chain_rule(n,farr,garr):
    r"""
    Evaluate the :math:`n`th derivative of :math:`f(g(x))` at x=0
    by specifying the Taylor series coefficients of the functions
    f and g.

    Arguments
    ---------
    farr : numpy array
        Taylor series coefficients of function f such that:
        f(u) = f[0] + f[1]*u + f[2]*u^2/2 + ... + f[n]*u^n/n! + ...
    garr : numpy array
        Taylor series coefficients of function g, such that:
        g(x) = g[0] + g[1]*x + g[2]*x^2/2 + ... + g[n]*x^n/n! + ...

    Returns
    -------
    float
        The value of :math:`\frac{d^n}{dx^n}f(g(x))\bigg|_{x=0}`
    """
    #
    return np.sum([farr[k] * bell(n,k,garr[1:n-k+2]) for k in range(1,n)])

def _to_uv_derivs(T,n,m,c,s):
    """
    Helper function for rotate_derivs_array
    """
    tot = np.zeros(2)
    R = np.array([[c,s],[-s,c]])
    for l in range(n+1):
        binom_nl = binom(n,l)
        for l1 in range(m+1):
            cfactor = c**(m+l-l1)
            sfactor = s**(n-l+l1)
            coeff = (-1)**l1 * binom_nl * binom(m,l1) * cfactor * sfactor
            tot +=  coeff * T[:,l+l1,n+m-l-l1]
    return R @ tot

def rotate_derivs_array(T,theta_rot):
    r"""
    Given the partial derivatives of a map, :math:`T(x,y)`, with respect
    to :math:`x` and :math:`y`, get the partial derivative with respect to new,
    rotated coordinates :math:`(u,v)` defined by

    .. math::
        \begin{pvec}
            u \\
            v
        \end{pvec}
        =
        R_\theta \cdot
        \begin{pvec}
            x \\
            y
        \end{pvec}

    where :math:`R_\theta` is a rotation matrix.
    """
    c,s = np.cos(theta_rot),np.sin(theta_rot)
    T1 = np.zeros(T.shape)
    for n in range(1,T.shape[1]):
        for l in range(n+1):
            T1[:,l,n-l] = _to_uv_derivs(T,l,n-l,c,s)
    return T1

### Utilities for calculating Psis ####
from collections import defaultdict
def deriv_of_coeff_and_pows(coeff,pows):
    new_pows = np.append(pows,0)
    results = []
    for m in range(len(pows)):
        if new_pows[m]>0:
            new_pows = np.append(pows,0)
            new_pows[m]-=1
            new_pows[m+1]+=1
            results.append((coeff * pows[m], new_pows))
    return results

def _consolidate_coeff_and_pows_list(cp_list):
    pows_arr = np.array([_[1] for _ in cp_list])
    coeff_arr = np.array([_[0] for _ in cp_list])
    i = 0
    tot = 0
    result = []
    while tot < len(coeff_arr):
        pows = pows_arr[i]
        msk = np.alltrue(pows_arr==pows,axis=1)
        coeff = np.sum(coeff_arr[msk])
        result+=[(coeff,pows)]
        # next index of unique powers
        tot += np.sum(msk)
        i=i+np.argmin(msk[i:])
    return result


def get_Psi_dicts(n):
    PsiOld = {
        (0,1):[(1,np.array([0,1]))],
        (1,0):[(1,np.array([0,0]))]
    }
    Psis = [0,PsiOld]
    for m in range(1,n):
        PsiNew = defaultdict(list)
        for ij,coeffs_and_pows in PsiOld.items():
            i,j = ij
            PsiNew[(i+1,j)] += coeffs_and_pows
            for coeff,pows in coeffs_and_pows:
                PsiNew[(i,j)] += deriv_of_coeff_and_pows(coeff,pows)
                add_one = pows.copy()
                add_one[1] +=1
                PsiNew[(i,j+1)] += [(coeff,add_one)]
        PsiOld = PsiNew.copy()
        for ij, lst in PsiOld.items():
            PsiOld[ij] = _consolidate_coeff_and_pows_list(lst)
        Psis.append(PsiOld)
    # Remove (0,1) entries from Psis
    [Psi.pop((0,1),None) for Psi in Psis[1:]]
    return Psis

def evaluate_Psi(Psi_dict,Tprimes_arr,farr):
    tot = 0
    for ij,coeffs_and_pows_list in Psi_dict.items():
        i,j = ij
        Tprime_ij = Tprimes_arr[i,j]
        sub_tot = 0
        for coeff,pows in coeffs_and_pows_list:
            if coeff==0:
                continue
            npows = pows.shape[0]
            sub_tot += coeff * np.product(farr[:npows]**pows)
        tot+= Tprime_ij * sub_tot
    return tot

def func_from_series(coeffs,x):
    """
    Given a set of Taylor series coefficients, (c_0,....,c_N), evalute
    the sum

    .. math::
        \sum_{n=0}^{N} c_n x^n / n!

    Arguments
    ---------
    coeffs : numpy array
        Values of Taylor series coeffieciens
    x : float
        Argument of function

    Returns
    -------
    float
    """
    return coeffs @ np.array([x**n/np.math.factorial(n) for n in range(coeffs.shape[0])])

def manifold_approx(u,n,farr,garr):
    f = lambda x: func_from_series(farr[:n+1],x)
    g = lambda x: func_from_series(garr[:n+1],x)
    p0 = np.array([u,f(u)])
    p1 = np.array([g(u),f(g(u))])
    return p0,p1

def solve_manifold_f_and_g(xunst,mapobj,Nmax,unstable=True):
    r"""
    Solve for Taylor series approximations of functions
    :math:`f` and :math:`g` satisfying

    .. math::
        T(x_u(s)) = x_* + g(s)\hat{x}_u + f(g(s))\hat{x}_\perp

    where ...

    Arguments
    ---------
    xunst : array-like, (2,)
        Unstable fixed point of map.
    mapobj : object
        A 2D map.
    Nmax : int
        Maximum order Taylor series coefficients to
        compute for f and g
    unstable : bool, optional
        If true, solve for the unstable manifold f and
        g functions, otherwise solve for the stable manifold
        functions.
    Returns
    -------
    R : ndarray, shape (2,2)
        Rotation matrix
    """
    # Array of partial derivatives up to order Nmax
    if unstable:
        T = mapobj.partial_derivs(xunst,Nmax)
    else:
        T = mapobj.inv_partial_derivs(xunst,Nmax)
    # jacobian evaluated at xunst
    jac = np.array([
        [T[0][1,0],T[0][0,1]],
        [T[1][1,0],T[1][0,1]]
    ])
    vals,vecs = np.linalg.eig(jac)
    isreal = np.logical_not(np.iscomplex(vals))
    assert np.alltrue(isreal), "Eigenvalues of map at point ({},{}) are\
    complex!".format(xunst[0],xunst[1])
    iunst = np.argmax(vals)
    # Unstable eigenvalue and direction
    lambdaU = vals[iunst]
    uvec = vecs[:,iunst]
    theta_rot = np.arctan2(uvec[1],uvec[0])
    # Rotation matrix
    s,c = np.sin(theta_rot), np.cos(theta_rot)
    R = np.array([[c,s],[-s,c]])

    Tprime = rotate_derivs_array(T,theta_rot)
    # Initialize iterative procedure to solve for 
    # f and g arrays
    TU_01 = Tprime[0][0,1]
    Tperp_01 = Tprime[1][0,1]
    farr,garr = np.zeros((2,Nmax+1))
    garr[1] = lambdaU
    Psis = get_Psi_dicts(Nmax)
    # Iteratively solve for f and g coeffs
    for n in range(2,Nmax + 1):
        Psi_dict=Psis[n]
        # Get numerical value of Psis
        PsiPerp = evaluate_Psi(Psi_dict,Tprime[1],farr)
        PsiU = evaluate_Psi(Psi_dict,Tprime[0],farr)
        denom = Tperp_01-lambdaU**n
        Bsum = _evaluate_chain_rule(n,farr,garr)
        farr[n] = (Bsum-PsiPerp)/denom
        garr[n] = TU_01 * farr[n] + PsiU
    return R, farr, garr
